#ifndef SCORER_H
#define SCORER_H

#include "ScoringParameters.h"
#include "Mseq.h"
#include "Profile.h"

template <class ScoreType>
class Scorer
{
	private:

	Mseq::size_type _n1, _n2;
	Vec < Vec < ScoreType > > _S;
	
	public:

	Scorer() : _n1(0), _n2(0) { initialize_prof_idx(); }

	Scorer(ScoringParameters<ScoreType> &sp)
	{
		_S = sp.get_substitution_matrix();
	}
	
	Scorer(Mseq::size_type n1, Mseq::size_type n2, ScoringParameters<ScoreType> &sp) : _n1(n1), _n2(n2)
	{
		if (!(sp.has_substitution_matrix()))
			throw BIO::E_MissingSubstMatrixEx();
		
		_S = sp.get_substitution_matrix();
	}

	void setN1(Mseq::size_type n1)
	{
		_n1 = n1;
	}

	void setN2(Mseq::size_type n2)
	{
		_n2 = n2;
	}

	void setScoringParameters(ScoringParameters<ScoreType> &sp)
	{
		_S = sp.get_substitution_matrix();
	}

	ScoreType scoreColumns(MIter& m1, MIter& m2)
	{
		//Count letter frequencies. Totally NUM_ALPHABET+1 letters = 5+1 letters = A, C, G, T, -, N
		unsigned short Fr1[BIO::NUM_ALPHABET + 1] = {0}, Fr2[BIO::NUM_ALPHABET + 1] = {0};
		ScoreType score = 0;

		for (MIter::size_type i = 0; i < _n1; ++i)
			++Fr1[CHAR_PROFILE_IDX[(unsigned char)(*(m1[i]))]];

		for (MIter::size_type j = 0; j < _n2; ++j)
		{
			++Fr2[CHAR_PROFILE_IDX[(unsigned char)(*(m2[j]))]];
		}
		
		/* Now score based on Sum-of-Pairs, without taking 'N's into account */
		for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
		{
			for (unsigned int j = 0; j < BIO::NUM_ALPHABET; ++j)
			{
				score += Fr1[i]*Fr2[j]*_S[i][j];
			}
		}
		return score;
	}

	ScoreType scoreColumn(MIter& m1, unsigned char letter)
	{
		//Count letter frequencies. Totally NUM_ALPHABET+1 letters = 5+1 letters = A, C, G, T, -, N
		unsigned short Fr1[BIO::NUM_ALPHABET + 1] = {0};
		ScoreType score = 0;

		for (MIter::size_type i = 0; i < _n1; ++i)
		{
			++Fr1[CHAR_PROFILE_IDX[(unsigned char)(*(m1[i]))]];
		}
		/* Now score based on Sum-of-Pairs, without taking 'N's into account */
		for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
		{
			score += Fr1[i]*_S[i][letter];
		}
		return score;
	}

	ScoreType scoreMIter(MIter &m)
	{
		//Count letter frequencies. Totally NUM_ALPHABET+1 letters = 5+1 letters = A, C, G, T, -, N
		unsigned short Fr[BIO::NUM_ALPHABET + 1] = {0};
		ScoreType score = 0;

		for (MIter::size_type i = 0; i < _n1; ++i)
			++Fr[CHAR_PROFILE_IDX[(unsigned char)(*(m[i]))]];


		/* Now score based on Sum-of-Pairs, without taking 'N's into account */
		for (unsigned int i = 0; i < BIO::NUM_ALPHABET; ++i)
		{
			for (unsigned int j = i; j < BIO::NUM_ALPHABET; ++j)
			{
				score += Fr[i]*Fr[j]*_S[i][j];
			}
		}
		return score;
	}
};

#endif
