#ifndef VEC_H
#define VEC_H

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                               *
 * Vec is a vector wrapper that does optional bounds checking    *
 *                                                               *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
* Notes:
* -----
*
* Define VEC_ENABLED_CHECKS if you actually want to do bounds
* checking. Production (release) versions should undefine that
* (hopefully... :)                             --George@2004-04-26
*
*/

#include <vector>

template< typename T, typename Alloc = std::allocator<T> >
class Vec : public std::vector<T, Alloc>
{
	typedef std::vector<T, Alloc> _Parent;

	public:

	typedef typename _Parent::size_type size_type;

	Vec() : _Parent() {}
	Vec(size_type size) : _Parent(size) {}
	Vec(size_type size, const T &value) : _Parent(size, value) {}
	Vec(const Vec &from) : _Parent(from) {}

#ifdef VEC_ENABLE_CHECKS
	/* "Safe" version of "[]" operator */
	T &operator[](size_type index)
	{
		/* Let "at" do the bound checking ;-) */
		return at(index);
	}

	/* "Safe" version of const "[]" operator */
	const T &operator[] (size_type index) const
	{
		/* Let "at" do the bound checking */
		return at(index);
	}
#endif
};
#endif
