#ifndef _localalignment_h_included_
#define _localalignment_h_included_

#include <iostream>
#include "System.h"
#include "LocalHit.h"
#include "Vec.h"

/**
 * The LocalAlgnment class represents full gapped alignments
 * It is represented as a list of ungapped alignments
 */

class LocalAlignment
{
 public:

	/* Constructors */
	/* ============ */

	/**
	 * Construct a local alignment centered at pos1, pos2.
	 */

	LocalAlignment(const LocalHit::size_type pos1, const LocalHit::size_type pos2) : _pos1(pos1), _pos2(pos2) {}

	/* Methods */
	/* ======= */

	/**
	 * Get the center of the alignment
	 */

	Seq::size_type getPos1() const { return _pos1; }
	Seq::size_type getPos2() const { return _pos2; }

	/**
	 * Get/Set the score of the alignent
	 */
	void setScore(const LocalHit::score_type score)	{	_score = score;	}
	LocalHit::score_type getScore() const	{	return _score; }

	/**
	 * Add another ungapped alignment to the end of the front
	 */
	void addToEndOfFront(const LocalHit& hit) { _preCoreChain.push_back(hit); }
	/**
	 * Add another ungapped alignment to the beginning of the back
	 */
	void addToBeginningOfBack(const LocalHit& hit)	{	_postCoreChain.push_back(hit); }

	/**
	 * This returns the alignment as a list of ungapped segments
	 */
	void getUngappedSegments(Vec<LocalHit>& segments)
		{
			segments.clear();

			if (_preCoreChain.size() > 0)
			{
				copy(_preCoreChain.begin(), _preCoreChain.end() - 1, back_inserter(segments));
			}
			//get the two middle ones
			if (_preCoreChain.size() == 0)
			{
				if (_postCoreChain.size() > 0)
				{
					LocalHit firstOfBack = *(_postCoreChain.end() - 1);
					segments.push_back(firstOfBack);
				}
				else
				{
					segments.push_back(LocalHit(_pos1, _pos2, 1, 0));
				}
			}
			else if (_postCoreChain.size() == 0)
			{

				if (_preCoreChain.size() > 0)
				{
					LocalHit endOfFront = *(_preCoreChain.end() - 1);
					segments.push_back(endOfFront);
				}
			}
			else 
			{
				LocalHit endOfFront = *(_preCoreChain.end() - 1);
				LocalHit firstOfBack = *(_postCoreChain.end() - 1);
				if (endOfFront.getDiagonal() == firstOfBack.getDiagonal())
				{
					LocalHit middle(firstOfBack);
					middle.extendLeft(firstOfBack.getPos1Begin() - endOfFront.getPos1Begin());
					segments.push_back(middle);
				}
				else
				{
					segments.push_back(endOfFront);
					segments.push_back(firstOfBack);
				}
			}
			if (_postCoreChain.size() > 0)
			{
				copy(_postCoreChain.rbegin() + 1, _postCoreChain.rend(), back_inserter(segments));
			}
		}

	/* Operators */
	/* ========= */

	bool operator<(const LocalAlignment& h2) const { return _pos1 < _pos2; }

 private:

	LocalHit::size_type _pos1;
	LocalHit::size_type _pos2;
	Vec<LocalHit> _preCoreChain;
	Vec<LocalHit> _postCoreChain;
	LocalHit::score_type _score;
};

namespace LocalAlignmentNS
{
	bool compare_scores(const LocalAlignment& a1, const LocalAlignment& a2);
}


#endif
