#include "MatchFinder.h"
#include "Profile.h"
#include <sstream>
#include <algorithm>
#include "Util.h"

void MatchFinder::setSequence1(Mseq& sequence1, ScoringParameters<score_type>& scoringParameters)
{
	_length1 = sequence1.getLength();
	_size1 = sequence1.getSize();

	_matrix = scoringParameters.get_substitution_matrix();
	_gapOpen = scoringParameters.get_gap_open();
	initialize_prof_idx();
	
	GAP_INDEX = CHAR_PROFILE_IDX[(unsigned char)GAP_CHAR];

	//This needs further exploration before we make this available

	if (_scoringType != ConsensusScoring)
	{
		Seq::size_type length = _length1;
		_spMatrix.clear();
		_spMatrix.resize(BIO::NUM_ALPHABET);
 		for (scoring_matrix::iterator it = _spMatrix.begin(); it != _spMatrix.end(); ++it)
 		{
			it->resize(length);
 		}


		MIter iter;
		sequence1.getMIBegin(iter);
		Scorer<score_type> scorer(scoringParameters);
		scorer.setN1(sequence1.getSize());
		for (unsigned int i = 0; i < length; ++i, ++iter)
		{
			for (unsigned int j = 0; j < BIO::NUM_ALPHABET; j++)
			{
				_spMatrix[j][i] = scorer.scoreColumn(iter, j);
			}
		}
	}
	
// 	if (_scoringType == PSIScoring)
// 	{
// 		Seq::size_type length = _length1;
// 		const double beta = 10;
// 		_psiMatrix.clear();
// 		_psiMatrix.resize(BIO::NUM_ALPHABET);
// 		_psiMatrixLength = length;
		
// 		for (scoring_matrix::iterator it = _psiMatrix.begin(); it != _psiMatrix.end(); ++it)
// 		{
// 			it->resize(length);
// 		}
// 		//now calculate the values
		
// 		//we need lambda
// 		_lambda = scoringParameters.getLambda();

// 		//calculate the qij
// 		Vec<Vec<double> > q;
// 		q.resize(NUM_NUCLEOTIDES);
// 		for (Vec<Vec<double> >::iterator it = q.begin(); it != q.end(); ++it)
// 		{
// 			it->resize(NUM_NUCLEOTIDES);
// 		}

// 		double sum = 0;
// 		for (Vec<Vec<double> >::size_type i = 0; i != NUM_NUCLEOTIDES; ++i)
// 		{
// 			for (Vec<Vec<double> >::size_type j = 0; j != NUM_NUCLEOTIDES; ++j)
// 			{
// 				q[i][j] = exp(static_cast<double>(_matrix[i][j] * _lambda)) / (NUM_NUCLEOTIDES * NUM_NUCLEOTIDES);
// 				sum += q[i][j];
// 			}
// 		}
// 		MIter it;
// 		sequence1.getMIBegin(it);

// 		for (Seq::size_type s = 0; s < length; ++s, ++it)
// 		{
// 			Seq::size_type numSeqs = 0;

// 			Vec<double> f(NUM_NUCLEOTIDES, 0);
// 			for (MIter::iterator iter = it.getBegin(); iter != it.getEnd(); iter++)
// 			{
// 				char letter = **iter;
// 				if (letter == ANY_CHAR)
// 				{
// 					letter = GAP_CHAR;
// 				}
// 				if (letter == GAP_CHAR)
// 				{
// 					continue;
// 				}
// 				numSeqs++;
// 				f[CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(letter))]]++;
// 			}
// 			const double alpha = numSeqs - 1;

// 			for (Vec<Vec<float> >::size_type i = 0; i != NUM_NUCLEOTIDES; ++i)
// 			{
// 				f[i] /= numSeqs;
// 			}
// 			for (Vec<Vec<float> >::size_type i = 0; i != NUM_NUCLEOTIDES; ++i)
// 			{
// 				double gi = 0;
// 				for (Vec<Vec<float> >::size_type j = 0; j != NUM_NUCLEOTIDES; ++j)
// 				{
// 					gi += q[i][j] * NUM_NUCLEOTIDES * f[j];
// 				}
// 				double Qi = (alpha * f[i] + beta * gi) / (alpha + beta);
// 				_psiMatrix[i][s] = static_cast<score_type>(log(Qi * NUM_NUCLEOTIDES) / _lambda);
// 			}
// 			for (Vec<Vec<float> >::size_type i = NUM_NUCLEOTIDES; i != BIO::NUM_ALPHABET; ++i)
// 			{
// 				for (Vec<Vec<float> >::size_type j = 0; j != NUM_NUCLEOTIDES; ++j)
// 				{
// 					_psiMatrix[i][s] += static_cast<score_type>(_matrix[i][j] * f[j]);
// 				}
// 			}
// 		}
//	}
//	else
	{
		MIter end;
		MIter it;
		sequence1.getMIEnd(end);
		sequence1.getMIBegin(it);
		_consensus1.resize(_length1);
		consensus_type::iterator cIt = _consensus1.begin();
		for (; it != end; ++it, ++cIt)
		{
			//treat N's like gaps
			char letter = it.getConsensusChar();
			if (letter == ANY_CHAR)
			{
				letter = GAP_CHAR;
			}
			*cIt = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(letter))];
		}
	}	

	_sequence1Name = sequence1.getSize() > 0 ? sequence1.getSequence(0).getTitle() : "Unkown";
}

void MatchFinder::setSequence2(Mseq& sequence2)
{
	_length2 = sequence2.getLength();
	_size2 = sequence2.getSize();
	_consensus2.clear();
	_consensus2.resize(_length2);
	MIter end;
	MIter it;
	sequence2.getMIEnd(end);
	sequence2.getMIBegin(it);
	consensus_type::iterator cIt = _consensus2.begin();
	for (; it != end; ++it, ++cIt)
	{
		//treat N's like gaps
		char letter = _size2 == 1 ? **(it.getBegin()) : it.getConsensusChar();
		if (letter == ANY_CHAR)
		{
			letter = GAP_CHAR;
		}

		*cIt = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(letter))];
	}
	_sequence2Name = sequence2.getSize() > 0 ? sequence2.getSequence(0).getTitle() : "Unkown";
}


void MatchFinder::initialize(ScoringParameters<score_type>& scoringParameters)												 
{
	_hitsToWrite.clear();
	_alignmentsToWrite.clear();
	_diagonalsToWrite.clear();
	//initialize last seeds
	_lastSeeds.clear();
	_fastHits.clear();
	//min is 1 - sequence2 size
	LocalHit::diagonal_type minDiagonal = 1 - _length1;
	_diagonalOffset = -1 * minDiagonal;
	LocalHit::diagonal_type maxDiagonal = _length2 - 1;

	_lastSeeds.resize(maxDiagonal - minDiagonal + 1, -1);
	_fastHits.resize(maxDiagonal - minDiagonal + 1, -1);

	_lambda = scoringParameters.getLambda();
	_K = scoringParameters.getK();


	_effectiveLength1 = _length1;
	_effectiveLength2 = _length2;

	if (!_bitFinalThreshold)
	{
		if (!_eValue)
		{
			_eValue = DEFAULT_E_VALUE;
		}

		_finalThreshold = static_cast<MatchFinder::score_type>(eToRaw(_eValue, _effectiveLength1, _effectiveLength2, _K, _lambda));
		_bitFinalThreshold = static_cast<MatchFinder::score_type>(rawToBit(_finalThreshold, _K, _lambda));
	}
	else
	{
		//convert from bit
		_finalThreshold = static_cast<MatchFinder::score_type>(bitToRaw(_bitFinalThreshold, _K, _lambda));

		//make sure e value syncs up
		_eValue = rawToE(_finalThreshold, _effectiveLength1, _effectiveLength2, _K, _lambda);
	}

	if (!_bitSemiFinalThreshold)
	{
		if (!_semiEValue)
		{
			_semiEValue = _eValue;
		}
		else if (_semiEValue < _eValue)
		{
			_semiEValue = _eValue;
		}
		_semiFinalThreshold = static_cast<MatchFinder::score_type>(eToRaw(_semiEValue, _effectiveLength1, _effectiveLength2, _K, _lambda));
		_bitSemiFinalThreshold = static_cast<MatchFinder::score_type>(rawToBit(_semiFinalThreshold, _K, _lambda));
	}
	else
	{
		if (_bitSemiFinalThreshold > _bitFinalThreshold)
		{
			_bitSemiFinalThreshold = _bitFinalThreshold;
		}
		//convert from bit
		_semiFinalThreshold = static_cast<MatchFinder::score_type>(bitToRaw(_bitSemiFinalThreshold, _K, _lambda));
		//make sure e value syncs up
		_semiEValue = rawToE(_semiFinalThreshold, _effectiveLength1, _effectiveLength2, _K, _lambda);
	}

	if (!_bitThresholdToDoGapped)
	{
		if (!_eValueToDoGapped)
		{
			_eValueToDoGapped = _eValue;
		}

		_thresholdToDoGapped = static_cast<MatchFinder::score_type>(eToRaw(_eValueToDoGapped, _effectiveLength1, _effectiveLength2, _K, _lambda));
		_bitThresholdToDoGapped = static_cast<MatchFinder::score_type>(rawToBit(_thresholdToDoGapped, _K, _lambda));
	}
	else
	{
		//convert from bit
		_thresholdToDoGapped = static_cast<MatchFinder::score_type>(bitToRaw(_bitThresholdToDoGapped, _K, _lambda));
		//make sure e value syncs up
		_eValueToDoGapped = rawToE(_thresholdToDoGapped, _effectiveLength1, _effectiveLength2, _K, _lambda);

	}

	if (!_bitCutoffThreshold)
	{
		_bitCutoffThreshold = DEFAULT_CUTOFF_THRESHOLD;
	}
	_cutoffThreshold = static_cast<MatchFinder::score_type>(bitToRaw(_bitCutoffThreshold, _K, _lambda));

	//WE HAVE DONE UP TO HERE

	if (!_bitSemiCutoffThreshold)
	{
		_bitSemiCutoffThreshold = _bitCutoffThreshold;
	}
	_semiCutoffThreshold = static_cast<MatchFinder::score_type>(bitToRaw(_bitSemiCutoffThreshold, _K, _lambda));

	if (!_bitGappedCutoff)
	{
		_bitGappedCutoff = DEFAULT_GAPPED_CUTOFF_THRESHOLD;
	}
	_gappedCutoff = static_cast<MatchFinder::score_type>(bitToRaw(_bitGappedCutoff, _K, _lambda));
}

void MatchFinder::writeOutResults(Mseq& sequence1, Mseq& sequence2, ostream& outStream, LocalHit::OutputType outputType, double indexTime, double scanTime)
{
	//outStream << "Seeds: " << _totalMatchesPreExtension << endl;
	//outStream << "Extensions: " << _numHitsExtended << endl;
	//outStream << "Matches: " << _totalMatches << endl;
	//outStream << "Total build time: " << indexTime << endl;
	//outStream << "Total scan time: " << scanTime << endl;

	unsigned int numFinalMatches = 0;

	Vec<LocalHit> hitsToBeWritten;
	set<LocalHit::diagonal_type>::const_iterator dEnd = _diagonalsToWrite.end();
	for (set<LocalHit::diagonal_type>::const_iterator dIt = _diagonalsToWrite.begin(); dIt != dEnd; ++dIt)
	{
		pair<hits_to_write_type::const_iterator, hits_to_write_type::const_iterator> eRange = _hitsToWrite.equal_range(*dIt);
		hits_to_write_type::const_iterator vBegin = eRange.first;
		hits_to_write_type::const_iterator vEnd = eRange.second;
		
		//get the total list
		Vec<LocalHit> currentHits;
		for (hits_to_write_type::const_iterator it = vBegin; it != vEnd; it++)
		{
			const LocalHit& hit = it->second;
			currentHits.push_back(hit);
		}

		sort(currentHits.begin(), currentHits.end());

		//we will now see if we can combine hits to achieve a higher necessary score,
		//but only if we aren't doing gapped extensions
			
		Vec<LocalHit>::iterator cEnd = currentHits.end();
//		unsigned int numInSum = 1;
		for (Vec<LocalHit>::iterator cIt = currentHits.begin(); cIt != cEnd; ++cIt)
		{
			Vec<LocalHit>::iterator cItNext = cIt + 1;
			
			//right now sum statistics aren't working
// 			if (cItNext != cEnd && _combine)
// 			{
// 				double normalizedScore = _lambda * cIt->getScore() - log(_K);
// 				//check if merging the next one adds to this score
// 				double combinedNormalizedScore = 0;
// 				if (numInSum < 2)
// 				{
// 					//here we add RAW, since we are beginning the sum
// 					combinedNormalizedScore = static_cast<LocalHit::score_type>(_lambda * (cIt->getScore() + cItNext->getScore()) - log(_K * _effectiveLength1 * _effectiveLength2) - (log(_K) + 2 * log(static_cast<double>(cItNext->getPos1Begin() - cIt->getPos2End()))) - M_LN2);
// 				}
// 				else
// 				{
// 					//here, we add to NORMALIZED, since updating a previously calculated sum (which is normal)
// 					combinedNormalizedScore = static_cast<LocalHit::score_type>(normalizedScore + _lambda * cItNext->getScore() - log(_K) - 2 * log(static_cast<double>(cItNext->getPos1Begin() - cIt->getPos2End())) + log(static_cast<double>(fact(numInSum))) - log(static_cast<double>(fact(numInSum + 1))));
// 				}
				
// 				if (combinedNormalizedScore > normalizedScore)
// 				{
// 					//merge the two hits and adjust the score
// 					cItNext->extendLeft(cItNext->getPos1Begin() - cIt->getPos1Begin());
// 					//convert back to raw
// 					cItNext->setScore(static_cast<LocalHit::score_type>((combinedNormalizedScore + log(_K)) / _lambda));
// 					numInSum++;
// 					//this one has been merged to the next one
// 					continue;
// 				}
// 				else
// 				{
// 					numInSum = 1;
// 				}
// 			}
			
			if (cIt->getScore() >= _finalThreshold)
			{
				numFinalMatches++;
				hitsToBeWritten.push_back(*cIt);
			}
		}
	}

	if (outputType == LocalHit::LEVEL2 || outputType == LocalHit::LEVEL3)
	{
		outStream << endl << ">" << _sequence2Name << endl << endl;
	}

	{
		//now get the strings of all of the alignments
		ostringstream hitsOutput;
		MatchFinder::statistics_type numWritten = 0;

		if (_doGapped)
		{
			//now write out all of the alignments
			//first, sort them

			sort(_alignmentsToWrite.begin(), _alignmentsToWrite.end(), &LocalAlignmentNS::compare_scores);

			for (Vec<LocalAlignment>::iterator it = _alignmentsToWrite.begin(); it != _alignmentsToWrite.end(); ++it)
			{
				//we know all alignments are unique based on criteria for constructing them
				numWritten += _writeOutAlignment(sequence1, sequence2, *it, outputType == LocalHit::LEVEL1 ? hitsOutput : outStream, outputType);
			}
		}
		
		sort(hitsToBeWritten.begin(), hitsToBeWritten.end(), &LocalHitNS::compare_scores);

		for (Vec<LocalHit>::iterator it = hitsToBeWritten.begin(); it != hitsToBeWritten.end(); ++it)
		{
			//make sure not to write out hits overlapping alignments
			//we don't check when we store them, only now
			LocalHit point = _getCenterPoint(*it);
			if (!_containedInAlignment(point))
			{
				numWritten += _writeOutHit(sequence1, sequence2, *it, outputType == LocalHit::LEVEL1 ? hitsOutput : outStream, outputType);
			}
		}
		if (outputType == LocalHit::LEVEL1)
		{
			outStream << ">" << _sequence2Name << endl;
			outStream << -1 << endl;
			outStream << numWritten << endl;
			outStream << hitsOutput.str() << endl;
		}
	}
}

void MatchFinder::_clearStats()
{
	_numHitsExtended = 0;
	_numHitsExtendedMore = 0;
	_numGappedExtensions = 0;
	_numCellsExplored = 0;
  _totalMatchesPreExtension = 0;
  _totalMatches = 0;
}

bool MatchFinder::_addHit(Mseq& sequence1, Seq::size_type position1, Mseq& sequence2, Seq::size_type position2, TyphonNS::Pattern::size_type seedLength)
{
	if (position1 + 1 > 601400 && position1 + 1 < 601526)
	{
		//cout << "Seed at Position " << position1 + 1 << "," << position2 + _start2Offset + 1 << endl;
	}

	LocalHit::diagonal_type diagonal = position2 - position1 + _diagonalOffset;

	LocalHit::diagonal_type hitP = _fastHits[diagonal];
	//this holds the old hit 

	if (hitP != NO_HIT)
	{
		Seq::size_type hitPos = static_cast<Seq::size_type>(hitP);
		if (hitPos >= position1)
		{
			//overlaps with a previously found high scoring hit
			return false;
		}
	}

	if (_maxDistance)
	{
		//see if we have a recent hit for this
		LocalHit::diagonal_type seedP = _lastSeeds[diagonal];
		if (seedP == NO_HIT)
		{
			//no hit on this diagonal; store this to maybe extend later
			_lastSeeds[diagonal] = position1 + seedLength - 1;
			return false;
		}
		
		Seq::size_type seedPos = static_cast<Seq::size_type>(seedP);

		//we only extend in this case if there is a recent hit on the diagonal
		if (seedPos >= position1)
		{
			//in this case the previous hit overlaps this
			//we ignore this hit
			return false;
		}
		if (seedPos + _maxDistance < position1)
		{
			//they're too far apart; replace it
			_lastSeeds[diagonal] = NO_HIT;
			_lastSeeds[diagonal] = position1 + seedLength - 1;
			return false;
		}
		else
		{
			//they match; let's extend their concatenation
			Seq::size_type difference = position1 - (seedPos - seedLength + 1);
			position1 -= difference;
			position2 -= difference;
			seedLength += difference;
			//erase it, so that next time we'll start fresh
			_lastSeeds[diagonal] = NO_HIT;
		}
	}

	//by now we know that we want to extend the hit
	LocalHit seed(position1, position2, seedLength, 0);

// 	if (_scoringType == PSIScoring)
// 	{
// 		if (!_extendHitPSI(seed))
// 		{
// 			return false;
// 		}
// 	}
// 	else
	{
		if (!_extendHitC(seed))
		{
			//this only happens if it's contained or there's a failure;
			//score doesn't matter
			return false;
		}
	}

	//at this point, it is unique, but may not be high scoring

	if (!_doGappedExtension(seed))
	{
		//didn't get gapped; check if its an HSP
		//we store all high scoring, even if overlap an alignment
		//check at end whether we need to remove dups
		if (seed.getScore() > _finalThreshold)
		{
			//we need to store this as an HSP
			_hitsToWrite.insert(make_pair(diagonal, seed));
			_diagonalsToWrite.insert(diagonal);
		}
	}
	else
	{
		//hit added in doGappedAlignment
	}

	if (seed.getScore() > _finalThreshold)
	{
		//every single HSP (whether gapped was done)
		//should be stored in fastHits
		//fastHits is just for *efficiency*, not correctness
		//we don't store the segments of the gapped alignment because checking them
		//is too hard, and not needed for correctness
		_fastHits[diagonal] = seed.getPos1End();
		return true;
	}
	else
	{
		return false;
	}
}

bool MatchFinder::_extendHitC(LocalHit& seed)
{
	_numHitsExtended++;

	Seq::size_type pos1Begin = seed.getPos1Begin();
	Seq::size_type pos2Begin = seed.getPos2Begin();
	//these are the end ITERATORS, so we need to add 1
	Seq::size_type pos1End = seed.getPos1End() + 1;

	//initial score of the seed
	score_type initialScore = 0;
	consensus_type::iterator absoluteBegin1 = _consensus1.begin();
	consensus_type::iterator absoluteBegin2 = _consensus2.begin();
	consensus_type::iterator end1 = absoluteBegin1 + pos1End;
	consensus_type::iterator it1 = absoluteBegin1 + pos1Begin;
	consensus_type::iterator it2 = absoluteBegin2 + pos2Begin;

	scoring_matrix::size_type spIt1 = pos1Begin;

	for (; it1 != end1; ++it1, ++it2, spIt1++)
	{
		if (_scoringType == ConsensusScoring)
		{
			initialScore += _matrix[*it1][*it2];
		}
		else
		{
			initialScore += _spMatrix[*it2][spIt1];
		}
	}
	seed.setScore(initialScore);

	if (!_extendHitC(seed, _semiCutoffThreshold))
	{
		return false;
	}
	else if (seed.getScore() < _semiFinalThreshold)
	{
		return false;
	}

	_numHitsExtendedMore++;

	if (_semiCutoffThreshold < _cutoffThreshold)
	{
		//maybe we can extend more
		_extendHitC(seed, _cutoffThreshold);
	}

	//always return true. Make sure to check the score above
	return true;

}

// bool MatchFinder::_extendHitPSI(LocalHit& seed)
// {
// 	_numHitsExtended++;

// 	Seq::size_type pos1Begin = seed.getPos1Begin();
// 	Seq::size_type pos2Begin = seed.getPos2Begin();
// 	//these are the end ITERATORS, so we need to add 1
// 	Seq::size_type pos1End = seed.getPos1End() + 1;

// 	//initial score of the seed
// 	score_type initialScore = 0;
// 	consensus_type::iterator absoluteBegin1 = _consensus1.begin();
// 	consensus_type::iterator absoluteBegin2 = _consensus2.begin();
// 	consensus_type::iterator end1 = absoluteBegin1 + pos1End;
// 	consensus_type::iterator it1 = absoluteBegin1 + pos1Begin;
// 	consensus_type::iterator it2 = absoluteBegin2 + pos2Begin;

// 	for (; it1 != end1; ++it1, ++it2)
// 	{
// 		initialScore += _matrix[*it1][*it2];
// 	}
// 	seed.setScore(initialScore);

// 	if (!_extendHitPSI(seed, _semiCutoffThreshold))
// 	{
// 		return false;
// 	}
// 	else if (seed.getScore() < _semiFinalThreshold)
// 	{
// 		return false;
// 	}

// 	_numHitsExtendedMore++;

// 	if (_semiCutoffThreshold < _cutoffThreshold)
// 	{
// 		//maybe we can extend more
// 		_extendHitPSI(seed, _cutoffThreshold);
// 	}

// 	return seed.getScore() >= _thresholdToDoGapped;

//}

bool MatchFinder::_extendHitC(LocalHit& seed, score_type cutoffThreshold)
{
	typedef LocalHit::size_type size_type;
	//store the properties of the seed for easier access
	size_type pos1Begin = seed.getPos1Begin();
	size_type pos2Begin = seed.getPos2Begin();
	//these are the end ITERATORS, so we need to add 1
	size_type pos1End = seed.getPos1End() + 1;
	size_type pos2End = seed.getPos2End() + 1;

	consensus_type::iterator absoluteBegin1 = _consensus1.begin();
	consensus_type::iterator absoluteBegin2 = _consensus2.begin();
	consensus_type::iterator absoluteEnd1 = _consensus1.end();
	consensus_type::iterator absoluteEnd2 = _consensus2.end();

	consensus_type::iterator it1 = absoluteBegin1 + pos1End;
	consensus_type::iterator it2 = absoluteBegin2 + pos2End;

	scoring_matrix::size_type spIt1 = pos1End;

	//initial score of the seed
	score_type initialScore = seed.getScore();

	//extend the hit to the right
	score_type bestScore = initialScore;
	score_type currentScore = initialScore;

	//stores how many past the best we go
	Seq::size_type numToUndo = 0;

	for (; it1 != absoluteEnd1 && it2 != absoluteEnd2; ++it1, ++it2, spIt1++)
	{
		score_type newScore = 0;
		if (_scoringType == ConsensusScoring)
		{
			newScore = _matrix[*it1][*it2];
		}
		else
		{
			newScore = _spMatrix[*it2][spIt1];
		}

		currentScore += newScore;
		seed.extendRight();

		if (currentScore >= bestScore)
		{
			bestScore = currentScore;
			numToUndo = 0;
		}
		else
		{
			numToUndo++;
		}
		
		if (currentScore < bestScore - cutoffThreshold)
		{
			break;
		}
	}
	//our score is the best one we found
	score_type rightScore = bestScore - initialScore;
	seed.undoExtendRight(numToUndo);

	bestScore = initialScore;
	currentScore = initialScore;
	//initialize since plus is destructive
	it1 = absoluteBegin1;
	it2 = absoluteBegin2;

	numToUndo = 0;

	for (it1 = it1 + pos1Begin - 1, it2 = it2 + pos2Begin - 1, spIt1 = pos1Begin - 1; it1 >= absoluteBegin1 && it2 >= absoluteBegin2; --it1, --it2, --spIt1)
	{
		//we know now that this position is not consumed by an existing seed; score it and extend
		score_type newScore = 0;
		if (_scoringType == ConsensusScoring)
		{
			newScore = _matrix[*it1][*it2];
		}
		else
		{
			newScore = _spMatrix[*it2][spIt1];
		}

		currentScore += newScore;
		seed.extendLeft();

		if (currentScore >= bestScore)
		{
			bestScore = currentScore;
			numToUndo = 0;
		}
		else
		{
			numToUndo++;
		}

		if (currentScore < bestScore - cutoffThreshold)
		{
			break;
		}
	}

	//now undo
	score_type leftScore = bestScore - initialScore;
	seed.undoExtendLeft(numToUndo);

	seed.setScore(initialScore + leftScore + rightScore);

	return true;
}


// bool MatchFinder::_extendHitPSI(LocalHit& seed, score_type cutoffThreshold)
// {
// 	_numHitsExtended++;
// 	typedef LocalHit::size_type size_type;
// 	//store the properties of the seed for easier access
// 	size_type pos1Begin = seed.getPos1Begin();
// 	size_type pos2Begin = seed.getPos2Begin();
// 	//these are the end ITERATORS, so we need to add 1
// 	size_type pos1End = seed.getPos1End() + 1;
// 	size_type pos2End = seed.getPos2End() + 1;

// 	//initial score of the seed
// 	typedef int it1_type;
// 	score_type initialScore = 0;
// 	it1_type absoluteBegin1 = 0;
// 	consensus_type::iterator absoluteBegin2 = _consensus2.begin();
// 	it1_type absoluteEnd1 = _psiMatrixLength;
// 	consensus_type::iterator absoluteEnd2 = _consensus2.end();

// 	it1_type end1 = absoluteBegin1 + pos1End;
// 	consensus_type::iterator end2 = absoluteBegin2 + pos2End;

// 	//we use an int because we may want to go to -1
// 	it1_type it1 = absoluteBegin1 + pos1Begin;
// 	consensus_type::iterator it2 = absoluteBegin2 + pos2Begin;

// 	for (; it1 != end1; ++it1, ++it2)
// 	{
// 		initialScore += _psiMatrix[*it2][it1];
// 	}
// 	//extend the hit to the right
// 	score_type bestScore = initialScore;
// 	score_type currentScore = initialScore;

// 	//stores how many past the best we go
// 	Seq::size_type numToUndo = 0;

// 	for (; it1 != absoluteEnd1 && it2 != absoluteEnd2; ++it1, ++it2)
// 	{
// 		score_type newScore = _psiMatrix[*it2][it1];
// 		currentScore += newScore;
// 		seed.extendRight();

// 		if (currentScore >= bestScore)
// 		{
// 			bestScore = currentScore;
// 			numToUndo = 0;
// 		}
// 		else
// 		{
// 			numToUndo++;
// 		}
		
// 		if (currentScore < bestScore - cutoffThreshold)
// 		{
// 			break;
// 		}
// 	}
// 	//our score is the best one we found
// 	score_type rightScore = bestScore - initialScore;
// 	seed.undoExtendRight(numToUndo);

// 	//extend the hit to the left
// 	it1_type previousEndInSequence1 = absoluteBegin1;
// 	bool haveHit = false;

// 	LocalHit::diagonal_type hitEnd = _fastHits[seed.getDiagonal() + _diagonalOffset];
// 	if (hitEnd != NO_HIT)
// 	{
// 		previousEndInSequence1 = (previousEndInSequence1 + 1) + static_cast<Seq::size_type>(hitEnd);
// 		haveHit = true;
// 	}

// 	bestScore = initialScore;
// 	currentScore = initialScore;
// 	//initialize since plus is destructive
// 	it1 = absoluteBegin1;
// 	it2 = absoluteBegin2;
// 	numToUndo = 0;
// 	for (it1 = it1 + pos1Begin - 1, it2 = it2 + pos2Begin - 1; it1 >= absoluteBegin1 && it2 >= absoluteBegin2; --it1, --it2)
// 	{

// 		//we know now that this position is not consumed by an existing seed; score it and extend
// 		score_type newScore = _psiMatrix[*it2][it1];

// 		currentScore += newScore;
// 		seed.extendLeft();

// 		if (currentScore >= bestScore)
// 		{
// 			bestScore = currentScore;
// 			numToUndo = 0;
// 		}
// 		else
// 		{
// 			numToUndo++;
// 		}
// 		if (currentScore < bestScore - cutoffThreshold)
// 		{
// 			break;
// 		}
// 	}
// 	//now undo
// 	score_type leftScore = bestScore - initialScore;
// 	seed.undoExtendLeft(numToUndo);

// 	seed.setScore(initialScore + leftScore + rightScore);

// 	return true;
// }

bool MatchFinder::_containedInAlignment(const LocalHit& hit)
{
	//determine if we don't already have a gapped alignment COMPLETELY including this one
	pair<hits_to_write_type::const_iterator, hits_to_write_type::const_iterator> eRange = _alignmentComponentsFound.equal_range(hit.getDiagonal());
	hits_to_write_type::const_iterator vBegin = eRange.first;
	hits_to_write_type::const_iterator vEnd = eRange.second;	
	
	for (hits_to_write_type::const_iterator vIt = vBegin; vIt != vEnd; ++vIt)
	{
		if (vIt->second.contains(hit))
		{
			return true;
		}
	}
	return false;
}

LocalHit MatchFinder::_getCenterPoint(const LocalHit& hit)
{
	LocalHit restrictedHit(0,0,0,0);
	if (hit.getLength() <= _hspSizeForGapped)
	{
		restrictedHit = hit;
	}
	else
	{
		consensus_type::iterator absoluteBegin1 = _consensus1.begin();
		consensus_type::iterator absoluteBegin2 = _consensus2.begin();
		consensus_type::iterator initialEnd1 = absoluteBegin1 + hit.getPos1Begin() + _hspSizeForGapped;

		consensus_type::iterator end1 = absoluteBegin1 + hit.getPos1End();
		consensus_type::iterator leadIt1 = absoluteBegin1 + hit.getPos1Begin();
		consensus_type::iterator leadIt2 = absoluteBegin2 + hit.getPos2Begin();
		consensus_type::iterator trailIt1 = absoluteBegin1 + hit.getPos1Begin();
		consensus_type::iterator trailIt2 = absoluteBegin2 + hit.getPos2Begin();

		LocalHit::score_type scoreOfWindow = 0;
		
		for (; leadIt1 != initialEnd1; ++leadIt1, ++leadIt2)
		{
			scoreOfWindow += _matrix[*leadIt1][*leadIt2];
		}

		Seq::size_type maxOffset = 0;
		LocalHit::score_type maxScoreOfWindow = scoreOfWindow;


		for (Seq::size_type curOffset = maxOffset + 1; leadIt1 != end1; ++leadIt1, ++leadIt2, ++trailIt1, ++trailIt2, ++curOffset)
		{
			scoreOfWindow += _matrix[*leadIt1][*leadIt2];
			scoreOfWindow -= _matrix[*trailIt1][*trailIt1];
			if (scoreOfWindow > maxScoreOfWindow)
			{
				maxScoreOfWindow = scoreOfWindow;
				maxOffset = curOffset;
			}
		}

		restrictedHit = LocalHit(hit.getPos1Begin() + maxOffset, hit.getPos2Begin() + maxOffset, _hspSizeForGapped, maxScoreOfWindow);

	}

	Seq::size_type amountToAdd = (restrictedHit.getLength() - 1) / 2;
	LocalHit point(restrictedHit.getPos1Begin() + amountToAdd, restrictedHit.getPos2Begin() + amountToAdd, 1, 0);
	return point;
}

bool MatchFinder::_doGappedExtension(LocalHit& hit)
{
	//first check the score
	if (!_doGapped || hit.getScore() < _thresholdToDoGapped)
	{
		return false;
	}

	//first restrict the HSP
	LocalHit point = _getCenterPoint(hit);
	
	if (_containedInAlignment(point))
	{
		//an alignment already completely contains this; no point in wasting time
		//(also guarantees that we won't have duplicate alignments)
		//we don't need to check HSPs because this is inside an HSP
		//which we have previously guaranteed does not overlap any other HSPs

		return false;
	}

	//get the score of this point
	LocalAlignment localAlignment(point.getPos1Begin(), point.getPos2Begin());
	localAlignment.setScore(_matrix[_consensus1[point.getPos1Begin()]][_consensus2[point.getPos2Begin()]]);

	_numGappedExtensions++;

	_doGappedExtension(localAlignment, FORWARD);
	_doGappedExtension(localAlignment, BACKWARD);

	if (localAlignment.getScore() > _finalThreshold)
	{
		//now we need to record the alignment for writing
		_alignmentsToWrite.push_back(localAlignment);

		//we also store all of the segments on the diagonal
		//this is to be used later when writing out ungapped hits to determine if we have an overlap
		//it will also be used to determine if we should bother doing an ungapped extension of a region
		Vec<LocalHit> segments;
		localAlignment.getUngappedSegments(segments);
		
		Vec<LocalHit>::const_iterator sEnd = segments.end();
		for (Vec<LocalHit>::const_iterator sIt = segments.begin(); sIt != sEnd; ++sIt)
		{
			_alignmentComponentsFound.insert(make_pair(sIt->getDiagonal(), *sIt));
		}
		return true;
	}
	else
	{
		return false;
	}
}

void MatchFinder::_doGappedExtension(LocalAlignment& alignment, GappedDirection direction)
{
	Seq::size_type pos1 = alignment.getPos1();
	Seq::size_type pos2 = alignment.getPos2();

	//first set up the iterators
	//the first sequence is Y; the second sequence is X
	consensus_type::size_type size1 = _consensus1.size();
	consensus_type::size_type size2 = _consensus2.size();
	consensus_type::size_type begin1 = pos1;
	consensus_type::size_type begin2 = pos2;

	int directionOfAdvance = direction == BACKWARD ? -1 : 1;

	//this is our DP matrix; it is stored by DIAGONAL
	typedef Vec<score_type> dp_column_type;
	typedef Vec<dp_column_type> dp_matrix_type;
	dp_matrix_type dpMatrixMatch;
	dp_matrix_type dpMatrixGapX;
	dp_matrix_type dpMatrixGapY;
	//we only store 3 diagonals
	dpMatrixMatch.resize(3);
	dpMatrixGapX.resize(3);
	dpMatrixGapY.resize(3);

	//this stores the maxY for each diagonal
	typedef int diag_offset_type;
	Vec<diag_offset_type> initY;
	Vec<diag_offset_type> initX;

	//this tells which vector is which
	dp_column_type::size_type diagCur = 0;
	dp_column_type::size_type diagBackOne = 2;
	dp_column_type::size_type diagBackTwo = 1;

	//this is our traceback matrix
	typedef Vec<AlignmentChar> traceback_column_type;
	typedef Vec<traceback_column_type> traceback_matrix_type;
	traceback_matrix_type tracebackMatrixMatch;
	traceback_matrix_type tracebackMatrixGapX;
	traceback_matrix_type tracebackMatrixGapY;

	//the maximal values
	score_type bestScore = 0;
	dp_column_type::size_type bestDiag = 0;
	dp_column_type::size_type bestD = 0;
	MatrixType bestMatrix = MATCH;

	//we do dp along the diagonal

	//these values store how much vertically and horizontally we can go
	//before going out of bounds
	dp_column_type::size_type maxY = 0;
	dp_column_type::size_type maxX = 0;

	if (direction == BACKWARD)
	{
		maxY = begin1;
		maxX = begin2;
	}
	else
	{
		maxY = size1 - begin1 - 1;
		maxX = size2 - begin2 - 1;
	}

	//initial case
	dpMatrixMatch[diagCur].resize(1, 0);
	dpMatrixGapX[diagCur].resize(1, -1 * _gapOpen);
	dpMatrixGapY[diagCur].resize(1, -1 * _gapOpen);

	tracebackMatrixMatch.resize(1);
	tracebackMatrixMatch[0].resize(1, ALIGNMENT_BEGIN);
	tracebackMatrixGapX.resize(1);
	tracebackMatrixGapX[0].resize(1, ALIGNMENT_BEGIN);
	tracebackMatrixGapY.resize(1);
	tracebackMatrixGapY[0].resize(1, ALIGNMENT_BEGIN);
	initY.resize(1, 0);
	initX.resize(1, 0);

	dp_column_type::size_type diagFromStart = 0;

	while (dpMatrixMatch[diagCur].size() != 0)
	{
		dp_column_type::size_type diagFromStartSize = dpMatrixMatch[diagCur].size();

		//the one two back is going to be the new cur
		dpMatrixMatch[diagBackTwo].clear();
		dpMatrixGapX[diagBackTwo].clear();
		dpMatrixGapY[diagBackTwo].clear();

		diag_offset_type newInitY = -1;
		diag_offset_type newInitX = -1;

		//calculate the beginning
		for (dp_column_type::size_type d = 0; d != diagFromStartSize; d++)
		{
			//check the current position
			if (dpMatrixMatch[diagCur][d] != MINF
					|| dpMatrixGapX[diagCur][d] != MINF
					|| dpMatrixGapY[diagCur][d] != MINF)
			{
				if (initY[diagFromStart] - d == maxY)
				{
					if (initX[diagFromStart] + d != maxX)
					{
						newInitY = initY[diagFromStart] - static_cast<diag_offset_type>(d);
						newInitX = initX[diagFromStart] + static_cast<diag_offset_type>(d) + 1;
					}
				}
				else
				{
					newInitY = initY[diagFromStart] + 1 - static_cast<diag_offset_type>(d);
					newInitX = initX[diagFromStart] + static_cast<diag_offset_type>(d);
				}
				break;
			}
			else if (d != 0)
			{
				//if we aren't all the way at the beginning, check over 1
				dp_column_type::size_type backD = d + initY[diagFromStart - 1] - initY[diagFromStart];
				if (dpMatrixMatch[diagBackOne][backD] != MINF
						|| dpMatrixGapX[diagBackOne][backD] != MINF
						|| dpMatrixGapY[diagBackOne][backD] != MINF)
				{
					newInitY = initY[diagFromStart] + 1 - static_cast<diag_offset_type>(d);
					newInitX = initX[diagFromStart] + static_cast<diag_offset_type>(d);
					break;
				}
			}
		}

		//found nothing
		if (newInitY == -1)
		{
			break;
		}

		//calculate the end
		dp_column_type::size_type nextDiagSize = 0;
		for (dp_column_type::size_type dCount = 0; dCount != diagFromStartSize; dCount++)
		{

			//get the position in the vector
			dp_column_type::size_type d = dpMatrixMatch[diagCur].size() - 1 - dCount;
			
			//always check the current position
			if (dpMatrixMatch[diagCur][d] != MINF
					|| dpMatrixGapX[diagCur][d] != MINF
					|| dpMatrixGapY[diagCur][d] != MINF)
			{
				//if the x is maxed out, actually do one less
				nextDiagSize = newInitY + d + 1 - initY[diagFromStart];
				if (initX[diagFromStart] + d == maxX)
				{
					nextDiagSize--;
				}
				break;
			}
			else if (d != 0)
			{
				//if we aren't all the way at the end, check over 1
				dp_column_type::size_type backD = d + initY[diagFromStart - 1] - initY[diagFromStart];
				if (dpMatrixMatch[diagBackOne][backD] != MINF
						|| dpMatrixGapX[diagBackOne][backD] != MINF
						|| dpMatrixGapY[diagBackOne][backD] != MINF)
				{
					nextDiagSize = newInitY - initY[diagFromStart] + static_cast<diag_offset_type>(d);
					break;
				}
			}
		}

		//now resize and rotate
		diagFromStart++;

		_numCellsExplored += nextDiagSize;

		dpMatrixMatch[diagBackTwo].resize(nextDiagSize);
		dpMatrixGapX[diagBackTwo].resize(nextDiagSize);
		dpMatrixGapY[diagBackTwo].resize(nextDiagSize);
		dp_column_type::size_type temp = diagBackTwo;
		diagBackTwo = diagBackOne;
		diagBackOne = diagCur;
		diagCur = temp;

		tracebackMatrixMatch.resize(tracebackMatrixMatch.size() + 1);
		tracebackMatrixMatch.back().resize(nextDiagSize);
		tracebackMatrixGapX.resize(tracebackMatrixGapX.size() + 1);
		tracebackMatrixGapX.back().resize(nextDiagSize);
		tracebackMatrixGapY.resize(tracebackMatrixGapY.size() + 1);
		tracebackMatrixGapY.back().resize(nextDiagSize);
		initY.resize(tracebackMatrixMatch.size());
		initY.back() = newInitY;
		initX.resize(tracebackMatrixMatch.size());
		initX.back() = newInitX;

		bool alwaysVert =  initX[diagFromStart] + dpMatrixMatch[diagCur].size() == initX[diagFromStart - 1] + dpMatrixMatch[diagBackOne].size();
		bool alwaysHor =  initY[diagFromStart] == initY[diagFromStart - 1];

		for (dp_column_type::size_type d = 0, y = initY[diagFromStart], x = initX[diagFromStart]; d < nextDiagSize; ++d, --y, ++x)
		{
			//check whether we have cells vert, hor, or diag
			bool haveVert = (d != dpMatrixMatch[diagCur].size() - 1 || alwaysVert);
			bool haveHor = (d != 0 || alwaysHor);
			bool haveDiag = haveVert && haveHor;

			score_type yScoreUsingGapX = MINF;
			score_type yScoreUsingGapY = MINF;
			score_type yScoreUsingMatch = MINF;
			score_type xScoreUsingGapX = MINF;
			score_type xScoreUsingGapY = MINF;
			score_type xScoreUsingMatch = MINF;
			score_type diagScoreUsingGapX = MINF;
			score_type diagScoreUsingGapY = MINF;
			score_type diagScoreUsingMatch = MINF;
			if (haveVert)
			{

				score_type fromScoreUsingGapY = dpMatrixGapY[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart] + 1];

				yScoreUsingGapY = fromScoreUsingGapY == MINF ? MINF : fromScoreUsingGapY + _matrix[GAP_INDEX][_consensus1[begin1 + y * directionOfAdvance]];

				score_type fromScoreUsingGapX = dpMatrixGapX[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart] + 1];
				yScoreUsingGapX = fromScoreUsingGapX == MINF ? MINF : fromScoreUsingGapX + _matrix[GAP_INDEX][_consensus1[begin1 + y * directionOfAdvance]] - _gapOpen;

				score_type fromScoreUsingMatch = dpMatrixMatch[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart] + 1];
				yScoreUsingMatch = fromScoreUsingMatch == MINF ? MINF : fromScoreUsingMatch + _matrix[GAP_INDEX][_consensus1[begin1 + y * directionOfAdvance]] - _gapOpen;

			}
			if (haveHor)
			{
				score_type fromScoreUsingGapX = dpMatrixGapX[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart]];
				xScoreUsingGapX = fromScoreUsingGapX == MINF ? MINF : fromScoreUsingGapX + _matrix[_consensus2[begin2 + x * directionOfAdvance]][GAP_INDEX];

				score_type fromScoreUsingGapY = dpMatrixGapY[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart]];
				xScoreUsingGapY = fromScoreUsingGapY == MINF ? MINF : fromScoreUsingGapY + _matrix[_consensus2[begin2 + x * directionOfAdvance]][GAP_INDEX] - _gapOpen;

				score_type fromScoreUsingMatch = dpMatrixMatch[diagBackOne][d + initY[diagFromStart - 1] - initY[diagFromStart]];
				xScoreUsingMatch = fromScoreUsingMatch == MINF ? MINF : fromScoreUsingMatch + _matrix[_consensus2[begin2 + x * directionOfAdvance]][GAP_INDEX] - _gapOpen;
				
			}
			if (haveDiag)
			{
				score_type fromScoreUsingGapX = dpMatrixGapX[diagBackTwo][d + initY[diagFromStart - 2] - initY[diagFromStart] + 1];
				diagScoreUsingGapX = fromScoreUsingGapX == MINF ? MINF : fromScoreUsingGapX + _matrix[_consensus2[begin2 + x * directionOfAdvance]][_consensus1[begin1 + y * directionOfAdvance]];				

				score_type fromScoreUsingGapY = dpMatrixGapY[diagBackTwo][d + initY[diagFromStart - 2] - initY[diagFromStart] + 1];
				diagScoreUsingGapY = fromScoreUsingGapY == MINF ? MINF : fromScoreUsingGapY + _matrix[_consensus2[begin2 + x * directionOfAdvance]][_consensus1[begin1 + y * directionOfAdvance]];				


				score_type fromScoreUsingMatch = dpMatrixMatch[diagBackTwo][d + initY[diagFromStart - 2] - initY[diagFromStart] + 1];
				diagScoreUsingMatch = fromScoreUsingMatch == MINF ? MINF : fromScoreUsingMatch + _matrix[_consensus2[begin2 + x * directionOfAdvance]][_consensus1[begin1 + y * directionOfAdvance]];				

			}

			//compute the max for match
			score_type maxScoreForMatch = diagScoreUsingMatch;
			AlignmentChar maxDirForMatch = DIAG_TO_MATCH;
			if (diagScoreUsingGapX > maxScoreForMatch)
			{
				maxScoreForMatch = diagScoreUsingGapX;
				maxDirForMatch = DIAG_TO_GAP_X;
			}
			if (diagScoreUsingGapY > maxScoreForMatch)
			{
				maxScoreForMatch = diagScoreUsingGapY;
				maxDirForMatch = DIAG_TO_GAP_Y;
			}

			//compute the max for gapY
			score_type maxScoreForGapY = yScoreUsingGapY;
			AlignmentChar maxDirForGapY = SEQUENCE1_TO_GAP_Y;
			if (yScoreUsingMatch > maxScoreForGapY)
			{
				maxScoreForGapY = yScoreUsingMatch;
				maxDirForGapY = SEQUENCE1_TO_MATCH;
			}
			if (yScoreUsingGapX > maxScoreForGapY)
			{
				maxScoreForGapY = yScoreUsingGapX;
				maxDirForGapY = SEQUENCE1_TO_GAP_X;
			}

			//compute the max for gapX
			score_type maxScoreForGapX = xScoreUsingGapX;
			AlignmentChar maxDirForGapX = SEQUENCE2_TO_GAP_X;


			if (xScoreUsingMatch > maxScoreForGapX)
			{
				maxScoreForGapX = xScoreUsingMatch;
				maxDirForGapX = SEQUENCE2_TO_MATCH;
			}
			if (xScoreUsingGapY > maxScoreForGapX)
			{
				maxScoreForGapX = xScoreUsingGapY;
				maxDirForGapX = SEQUENCE2_TO_GAP_Y;
			}

			//now update the best score
			//for match
			if (maxScoreForMatch < bestScore - _gappedCutoff)
			{
				maxScoreForMatch = MINF;
				maxDirForMatch = NO_VALUE;
			}
			else if (maxScoreForMatch > bestScore)
			{
				bestScore = maxScoreForMatch;
				bestDiag = diagFromStart;
				bestD = d;
				bestMatrix = MATCH;
			}

			//for gapX
			if (maxScoreForGapX < bestScore - _gappedCutoff)
			{
				maxScoreForGapX = MINF;
				maxDirForGapX = NO_VALUE;
			}
			else if (maxScoreForGapX > bestScore)
			{
				bestScore = maxScoreForGapX;
				bestDiag = diagFromStart;
				bestD = d;
				bestMatrix = GAP_X;
			}

			//for gapY
			if (maxScoreForGapY < bestScore - _gappedCutoff)
			{
				maxScoreForGapY = MINF;
				maxDirForGapY = NO_VALUE;
			}
			else if (maxScoreForGapY > bestScore)
			{
				bestScore = maxScoreForGapY;
				bestDiag = diagFromStart;
				bestD = d;
				bestMatrix = GAP_Y;
			}

			//store values in the matrices
			dpMatrixMatch[diagCur][d] = maxScoreForMatch;
			dpMatrixGapX[diagCur][d] = maxScoreForGapX;
			dpMatrixGapY[diagCur][d] = maxScoreForGapY;

			tracebackMatrixMatch.back()[d] = maxDirForMatch;
			tracebackMatrixGapX.back()[d] = maxDirForGapX;
			tracebackMatrixGapY.back()[d] = maxDirForGapY;

		}
	}

	//now we need to determine the alignment
	dp_column_type::size_type diag = bestDiag;
	dp_column_type::size_type d = bestD;
	MatrixType mat = bestMatrix;

	LocalHit currentHit(0,0,0,0);

	bool done = false;
	bool haveHit = false;

	while (!done)
	{
		bool needToPush = false;
		traceback_matrix_type& curTracebackMatrix = (mat == MATCH ? tracebackMatrixMatch : (mat == GAP_X ? tracebackMatrixGapX : tracebackMatrixGapY));

		AlignmentChar curDir = curTracebackMatrix[diag][d];

		switch (curDir)
		{
			case DIAG_TO_MATCH :
			case DIAG_TO_GAP_X :
			case DIAG_TO_GAP_Y :
			{

				if (!haveHit)
				{
					currentHit = LocalHit(begin1 + (initY[diag] - d) * directionOfAdvance, begin2 + (initX[diag] + d) * directionOfAdvance, 1, 0);
					haveHit = true;
				}
				else
				{
					if (direction == BACKWARD)
					{
						currentHit.extendRight(1);
					}
					else
					{
						currentHit.extendLeft(1);
					}
				}

				d += static_cast<dp_column_type::size_type>(initY[diag - 2] - initY[diag] + 1);
				diag -= 2;

				mat = curDir == DIAG_TO_MATCH ? MATCH : (curDir == DIAG_TO_GAP_X ? GAP_X : GAP_Y);
				break;
			}
			case SEQUENCE1_TO_MATCH :
			case SEQUENCE1_TO_GAP_X :
			case SEQUENCE1_TO_GAP_Y :
			{

				d += static_cast<dp_column_type::size_type>(initY[diag - 1] - initY[diag] + 1);
				diag--;

				needToPush = haveHit;
				mat = curDir == SEQUENCE1_TO_MATCH ? MATCH : (curDir == SEQUENCE1_TO_GAP_X ? GAP_X : GAP_Y);
				break;
			}
			case SEQUENCE2_TO_MATCH :
			case SEQUENCE2_TO_GAP_X :
			case SEQUENCE2_TO_GAP_Y :
			{
				d += static_cast<dp_column_type::size_type>(initY[diag - 1] - initY[diag]);
				diag--;
				needToPush = haveHit;
				mat = curDir == SEQUENCE2_TO_MATCH ? MATCH : (curDir == SEQUENCE2_TO_GAP_X ? GAP_X : GAP_Y);
				break;
			}
			case ALIGNMENT_BEGIN :
			case NO_VALUE :
			default :
			{
				done = true;
			}

		}
		if (needToPush)
		{
			haveHit = false;
			if (direction == BACKWARD)
			{
				alignment.addToEndOfFront(currentHit);
			}
			else
			{
				alignment.addToBeginningOfBack(currentHit);
			}

		}
	}
	//get the last one
	if (haveHit)
	{
		if (direction == BACKWARD)
		{
			alignment.addToEndOfFront(currentHit);
		}
		else
		{
			alignment.addToBeginningOfBack(currentHit);
		}
	}
	alignment.setScore(alignment.getScore() + bestScore);

}

//the following two methods need to be redone

MatchFinder::statistics_type MatchFinder::_writeOutHit(Mseq& sequence1, Mseq& sequence2, LocalHit hit, ostream& outStream, LocalHit::OutputType outputType, bool recalculateScore) const
{
	MatchFinder::statistics_type numWritten = 0;
	if (outputType == LocalHit::LEVEL1 || outputType == LocalHit::LEVEL2)
	{
		//we have hits for chaining
		Vec<LocalHit> hitsToWriteHere;

		//check if we have to split
		Seq::size_type hitLength = hit.getLength();
		if ((_maxHitLength && hitLength > _maxHitLength) || recalculateScore)
		{
			//break it up
			Seq::size_type newHitLength = _maxHitLength == 0 ? hitLength : hitLength / static_cast<unsigned int>(ceil(static_cast<float>(hitLength) / _maxHitLength)); 

			Seq::size_type hitEnd1 = hit.getPos1End();

			Seq::size_type startNewHit1 = hit.getPos1Begin();
			Seq::size_type startNewHit2 = hit.getPos2Begin();

			while(startNewHit1 <= hitEnd1)
			{
				Seq::size_type actualNewHitLength = min(newHitLength, hitEnd1 - startNewHit1 + 1);

				LocalHit newHit(startNewHit1, startNewHit2, actualNewHitLength, 0);

				//avoid having a really small hit next time
				Seq::size_type nextHitLength = hitEnd1 - startNewHit1 - actualNewHitLength + 1;
				if (nextHitLength < newHitLength)
				{
					actualNewHitLength += nextHitLength;
					newHit = LocalHit(startNewHit1, startNewHit2, actualNewHitLength, 0);
				}

				//score this segment
				LocalHit::size_type pos1Begin = newHit.getPos1Begin();
				LocalHit::size_type pos2Begin = newHit.getPos2Begin();
				//these are the end ITERATORS, so we need to add 1
				LocalHit::size_type pos1End = newHit.getPos1End() + 1;

				score_type initialScore = 0;
				//if (_scoringType == PSIScoring)
// 					{
// 						//initial score of the seed
// 						Seq::size_type absoluteBegin1 = 0;
// 						consensus_type::const_iterator absoluteBegin2 = _consensus2.begin();

// 						Seq::size_type end1 = absoluteBegin1 + pos1End;
// 						Seq::size_type it1 = absoluteBegin1 + pos1Begin;
// 						consensus_type::const_iterator it2 = absoluteBegin2 + pos2Begin;

// 						for (; it1 != end1; ++it1, ++it2)
// 						{
// 							initialScore += _psiMatrix[*it2][it1];
// 						}
// 					}
//					else
				{
					//initial score of the seed
					consensus_type::const_iterator absoluteBegin1 = _consensus1.begin();
					consensus_type::const_iterator absoluteBegin2 = _consensus2.begin();

					consensus_type::const_iterator end1 = absoluteBegin1 + pos1End;
					consensus_type::const_iterator it1 = absoluteBegin1 + pos1Begin;
					consensus_type::const_iterator it2 = absoluteBegin2 + pos2Begin;
					for (; it1 != end1; ++it1, ++it2)
					{
						initialScore += _matrix[*it1][*it2];
					}
				}
				newHit.setScore(initialScore);

				//update for the next loop iteration
				startNewHit1 = newHit.getPos1End() + 1;
				startNewHit2 = newHit.getPos2End() + 1;

				hitsToWriteHere.push_back(newHit);
			}
		}
		else
		{
			hitsToWriteHere.push_back(hit);
		}

		if (outputType == LocalHit::LEVEL2)
		{
			outStream << "a {" << endl << "\ts " << hit.getScore() << endl;
			Vec<LocalHit>::iterator bIt = hitsToWriteHere.begin();
			bIt->writeBoundary(true, _start1Offset, _start2Offset, outStream);
			Vec<LocalHit>::reverse_iterator eIt = hitsToWriteHere.rbegin();
			eIt->writeBoundary(false, _start1Offset, _start2Offset, outStream);
		}
		for (Vec<LocalHit>::iterator it = hitsToWriteHere.begin(); it != hitsToWriteHere.end(); ++it)
		{
			it->writeOut(outputType, outStream, sequence1, sequence2, _start1Offset, _start2Offset, _matrix);
			if (outputType != LocalHit::LEVEL2)
			{
				outStream << endl;
			}
			numWritten++;
		}
		
		if (outputType == LocalHit::LEVEL2)
		{
			outStream << "}" << endl;
		}

	}
	else if (outputType == LocalHit::LEVEL3)
	{
		score_type bitScore = static_cast<score_type>(rawToBit(hit.getScore(), _K, _lambda));
		double e = rawToE(hit.getScore(), _effectiveLength1, _effectiveLength2, _K, _lambda);
		outStream << " Score = " << hit.getScore() << " (" << bitScore << " bits), Expect = " << e << endl;
		Seq::size_type numIdentities = hit.getNumberOfIdentities(sequence1, sequence2, _matrix);
		Seq::size_type numPositive = hit.getNumberOfIdentities(sequence1, sequence2, _matrix, false);
		
		outStream << " Identities = " << numIdentities << "/" << hit.getLength() << " (" << numIdentities * 100 / hit.getLength() << "%), Positives = " << numPositive << "/" << hit.getLength() << " (" << numPositive * 100 / hit.getLength() << "%), Strand = Plus / Plus" << endl;

		MIter it1;
		sequence1.getMIBegin(it1);
		it1 = it1 + hit.getPos1Begin();
		Seq::size_type numToPrint = hit.getLength();
		MIter it2;
		sequence2.getMIBegin(it2);
		it2 = it2 + hit.getPos2Begin();

		string sequence1String = "";
		string sequence2String = "";
		string matchString = "";

		ostringstream data;

		unsigned int lineNum = 0;
		unsigned int numPrintedThisLine = 0;

		Seq::size_type lengthPrinted = 0;
		
		for (; lengthPrinted < numToPrint; ++it1, ++it2, lengthPrinted++, numPrintedThisLine++)
		{
			if (lengthPrinted % MAX_PER_LINE == 0 && lengthPrinted > 0)
			{
				_printBlastLine(hit.getPos2Begin() + LocalHit::OFFSET + lineNum * MAX_PER_LINE, hit.getPos2Begin() + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, hit.getPos1Begin() + LocalHit::OFFSET + lineNum * MAX_PER_LINE, hit.getPos1Begin() + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);
				sequence1String = "";
				sequence2String = "";
				matchString = "";
				lineNum++;
				numPrintedThisLine = 0;
			}
			sequence1String = sequence1String + it1.getConsensusChar();
			sequence2String = sequence2String + it2.getConsensusChar();
			if (it1.getConsensusChar() == it2.getConsensusChar())
			{
				matchString = matchString + "|";
			}
			else
			{
				matchString = matchString + " ";
			}
		}

		_printBlastLine(hit.getPos2Begin() + LocalHit::OFFSET + lineNum * MAX_PER_LINE, hit.getPos2Begin() + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, hit.getPos1Begin() + LocalHit::OFFSET + lineNum * MAX_PER_LINE, hit.getPos1Begin() + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);
		
		outStream << data.str();
		return 1;
	}
	return numWritten;
}

void MatchFinder::_printBlastLine(Seq::size_type queryStartPos, Seq::size_type queryEndPos, const Seq::size_type queryOff, const string& queryString, Seq::size_type sbjctStartPos, Seq::size_type sbjctEndPos, const Seq::size_type sbjctOff, const string& sbjctString, const string& matchString, ostream& outStream) const
{
	queryStartPos += queryOff;
	queryEndPos += queryOff;
	sbjctStartPos += sbjctOff;
	sbjctEndPos += sbjctOff;

	ostringstream qs;
	qs << queryStartPos;
	string queryStart(qs.str());
	ostringstream ss;
	ss << sbjctStartPos;
	string sbjctStart(ss.str());

	string::size_type indentation = (queryStart.size() > sbjctStart.size() ? queryStart.size() : sbjctStart.size()) + 1;
	outStream << endl << "Query: " << queryStart;
	for (string::size_type i = 0; i < indentation - queryStart.size(); i++)
	{
		outStream << " ";
	}
	outStream << queryString << " " << queryEndPos << endl;
	for (string::size_type i = 0; i < indentation + 7; i++)
	{
		outStream << " ";
	}				
	outStream << matchString << endl;
	outStream << "Sbjct: " <<  sbjctStart;
	for (string::size_type i = 0; i < indentation - sbjctStart.size(); i++)
	{
		outStream << " ";
	}
	outStream << sbjctString << " " << sbjctEndPos << endl << endl;
}

MatchFinder::statistics_type MatchFinder::_writeOutAlignment(Mseq& sequence1, Mseq& sequence2, LocalAlignment alignment, ostream& outStream, LocalHit::OutputType outputType) const
{
	MatchFinder::statistics_type numWritten = 0;

	Vec<LocalHit> segments;
	alignment.getUngappedSegments(segments);

	if (outputType == LocalHit::LEVEL1)
	{
		Vec<LocalHit>::iterator end = segments.end();
		for (Vec<LocalHit>::iterator it = segments.begin(); it != end; ++it)
		{
			numWritten += _writeOutHit(sequence1, sequence2, *it, outStream, outputType, true);
		}
	}
	else if (outputType == LocalHit::LEVEL2)
	{
		outStream << "a {" << endl << "\ts " << alignment.getScore() << endl;
		Vec<LocalHit>::iterator bIt = segments.begin();
		bIt->writeBoundary(true, _start1Offset, _start2Offset, outStream);
		Vec<LocalHit>::reverse_iterator eIt = segments.rbegin();
		eIt->writeBoundary(false, _start1Offset, _start2Offset, outStream);
	
		for (Vec<LocalHit>::iterator it = segments.begin(); it != segments.end(); ++it)
		{
			it->writeOut(outputType, outStream, sequence1, sequence2, _start1Offset, _start2Offset, _matrix);
			numWritten++;
		}
		
		outStream << "}" << endl;
	}
	else if (outputType == LocalHit::LEVEL3)
	{

		score_type bitScore = static_cast<score_type>(rawToBit(alignment.getScore(), _K, _lambda));
		double e = rawToE(alignment.getScore(), _effectiveLength1, _effectiveLength2, _K, _lambda);
		outStream << " Score = " << alignment.getScore() << " (" << bitScore << " bits), Expect = " << e << endl;

		numWritten = segments.size();

		string sequence1String = "";
		string sequence2String = "";
		string matchString = "";

		ostringstream data;

		Seq::size_type numTotal = 0;
		Seq::size_type numIdentities = 0;
		Seq::size_type numPositive = 0;

		unsigned int lineNum = 0;
		unsigned int numPrintedThisLine = 0;

		Seq::size_type totalLengthPrinted = 0;

		Seq::size_type beginPos1 = segments.begin()->getPos1Begin() + LocalHit::OFFSET;
		Seq::size_type beginPos2 = segments.begin()->getPos2Begin() + LocalHit::OFFSET;

		for (Vec<LocalHit>::iterator it = segments.begin(); it != segments.end(); ++it)
		{
			MIter it1;
			sequence1.getMIBegin(it1);
			it1 = it1 + it->getPos1Begin();
			Seq::size_type numToPrint = it->getLength();
			numTotal += numToPrint;

			MIter it2;
			sequence2.getMIBegin(it2);
			it2 = it2 + it->getPos2Begin();

			Seq::size_type lengthPrinted = 0;

			for (; lengthPrinted < numToPrint; ++it1, ++it2, lengthPrinted++, numPrintedThisLine++, totalLengthPrinted++)
			{
				if (totalLengthPrinted % MAX_PER_LINE == 0 && totalLengthPrinted > 0)
				{
					_printBlastLine(beginPos2 + lineNum * MAX_PER_LINE, beginPos2 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, beginPos1 + LocalHit::OFFSET + lineNum * MAX_PER_LINE, beginPos1 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);
					sequence1String = "";
					sequence2String = "";
					matchString = "";
					lineNum++;
					numPrintedThisLine = 0;
				}

				char c1 = it1.getConsensusChar();
				char c2 = it2.getConsensusChar();
				unsigned int pos1 = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(c1))];
				unsigned int pos2 = CHAR_PROFILE_IDX[static_cast<int>(static_cast<unsigned char>(c2))];

				if (c1 == c2)
				{
					numIdentities++;
					matchString += "|";
				}
				else
				{
					matchString += " ";
				}

				if (_matrix[pos1][pos2] > 0)
				{
					numPositive++;
				}

				sequence1String += c1;
				sequence2String += c2;
			}

			if (it != segments.end() - 1)
			{
				Vec<LocalHit>::iterator nextIt = it + 1;

				numToPrint = nextIt->getPos1Begin() - it->getPos1End() - 1;
				lengthPrinted = 0;

				for (; lengthPrinted < numToPrint; ++it1, lengthPrinted++, numPrintedThisLine++, totalLengthPrinted++)
				{
					if (totalLengthPrinted % MAX_PER_LINE == 0 && totalLengthPrinted > 0)
					{
						_printBlastLine(beginPos2 + 1 + lineNum * MAX_PER_LINE, beginPos2 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, beginPos1 + 1 + lineNum * MAX_PER_LINE, beginPos1 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);

						sequence1String = "";
						sequence2String = "";
						matchString = "";
						lineNum++;
						numPrintedThisLine = 0;
					}

					char c1 = it1.getConsensusChar();
					matchString += " ";
					sequence1String += c1;
					sequence2String += "-";
				}

				numToPrint = nextIt->getPos2Begin() - it->getPos2End() - 1;
				lengthPrinted = 0;

				for (; lengthPrinted < numToPrint; ++it2, lengthPrinted++, numPrintedThisLine++, totalLengthPrinted++)
				{
					if (totalLengthPrinted % MAX_PER_LINE == 0 && totalLengthPrinted > 0)
					{
						_printBlastLine(beginPos2 + 1 + lineNum * MAX_PER_LINE, beginPos2 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, beginPos1 + 1 + lineNum * MAX_PER_LINE, beginPos1 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);

						sequence1String = "";
						sequence2String = "";
						matchString = "";
						lineNum++;
						numPrintedThisLine = 0;
					}

					char c2 = it2.getConsensusChar();
					matchString += " ";
					sequence2String += c2;
					sequence1String += "-";
				}
			}
		}

		_printBlastLine(beginPos2 + 1 + lineNum * MAX_PER_LINE, beginPos2 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start2Offset, sequence2String, beginPos1 + 1 + lineNum * MAX_PER_LINE, beginPos1 + lineNum * MAX_PER_LINE + numPrintedThisLine, _start1Offset, sequence1String, matchString, data);

		outStream << " Identities = " << numIdentities << "/" << numTotal << " (" << numIdentities * 100 / numTotal << "%), Positives = " << numPositive << "/" << numTotal << " (" << numPositive * 100 / numTotal << "%), Strand = Plus / Plus" << endl;

		outStream << data.str();

	}
	return numWritten;
}



