#ifndef _nextpatternsetgenerator_h_included_
#define _nextpatternsetgenerator_h_included_

#include "System.h"
#include "RegionList.h"
#include "Region.h"
#include "Pattern.h"
#include "ProbVector.h"
#include "TyphonMarkovModel.h"
#include "Vec.h"


class NextPatternSetGenerator
{
 public:

	/**
	 * Given a multiple alignment and a region list,
	 * perform any needed initialization. This will 
	 * be called one time at the beginning
	 */

	virtual ~NextPatternSetGenerator() {};

	void initialize(const ProbVector& pv, const RegionList& r, const Vec<TyphonNS::Pattern>& candidatePatterns, const unsigned int budget, const unsigned int order)
		{ 
			_remainingBudget = budget;
			copy(candidatePatterns.begin(), candidatePatterns.end(), back_inserter(_candidatePatterns));
			_initialize(pv, r, budget, order);
		}

	/**
	 * Return whether or not there is another pattern set we can fetch
	 */
	virtual bool hasMoreBudget() { return _remainingBudget; }

	/**
	 * Fetch the next region and pattern set
	 */
	virtual void nextPatternSet(Region& r, Vec<Vec<TyphonNS::Pattern>::size_type>& p) = 0;


 protected:

	virtual void _initialize(const ProbVector& pv, const RegionList& r, const unsigned int budget, const unsigned int order) = 0;

	void _finished() { _remainingBudget = 0; }
	const Vec<TyphonNS::Pattern>& _getPatterns() const { return _candidatePatterns; }
	TyphonNS::Pattern::size_type _getSpan() const { return _candidatePatterns.size() ? _candidatePatterns[0].getPatternLength() : 0; }
	TyphonNS::Pattern::size_type _getWeight() const { return _candidatePatterns.size() ? _candidatePatterns[0].getKLength() : 0; }

	void _subtractFromBudget(unsigned int curBudget)
		{
			_remainingBudget = (curBudget > _remainingBudget ? 0 : _remainingBudget - curBudget);
		}

	unsigned int _remainingBudget;
 private:

	Vec<TyphonNS::Pattern> _candidatePatterns;

};

#endif
