#include "NextPatternSetGeneratorByRegionClass.h"
#include "MandalaWrapper.h"
#include <list>

void NextPatternSetGeneratorByRegionClass::nextPatternSet(Region& r, Vec<Vec<TyphonNS::Pattern>::size_type>& p)
{
	const Region& nextRegion = _regionList.getRegion(_curRegion);
	r.setStart(nextRegion.getStart());
	r.setEnd(nextRegion.getEnd());

	region_attr_type curType = _regionList.getRegionType(_curRegion);

	while (_curRegion < _regionList.getNumRegions() - 1 && _regionList.getRegionType(_curRegion + 1) == curType)
	{
		_curRegion++;
		r.setEnd(_regionList.getRegion(_curRegion).getEnd());
	}

	Vec<Vec<TyphonNS::Pattern>::size_type>& patterns = _regionClassToPatternSubset[_regionList.getRegionType(_curRegion)];
	p.clear();
	p.resize(patterns.size());
	copy(patterns.begin(), patterns.end(), p.begin());
	_curRegion++;

	//update the budget
	if (_curRegion >= _regionList.getNumRegions())
	{
		_finished();
	}
	else
	{
		_subtractFromBudget(r.getLength());
	}

	TyphonNS::Pattern::size_type spanOverlap = _getSpan() / 2;
	r.setEnd(r.getEnd() + spanOverlap - 1);
	if (r.getEnd() > _endPoint)
	{
		r.setEnd(_endPoint);
	}
	if (r.getStart() > spanOverlap)
	{
		r.setStart(r.getStart() - spanOverlap);
	}

}

void NextPatternSetGeneratorByRegionClass::_initialize(const ProbVector& pv, const RegionList& regionList, const unsigned int budget, const unsigned int order)
{
	_curRegion = 0;
	_regionList = regionList;

	_regionList.breakRegions(MAX_REGION_LENGTH);

	//just to ensure that we have the overlap correct
	_endPoint = _regionList.getRegion(_regionList.getNumRegions() - 1).getEnd();

	//store a vector of the region lengths we'll use
	double increment = MAX_REGION_LENGTH_FOR_MULTIPLE / NUM_REGION_LENGTHS;
	Vec<unsigned int> regionLengths;
	for (unsigned int i = 0; i < NUM_REGION_LENGTHS; ++i)
	{
		regionLengths.push_back(static_cast<unsigned int>((i + 1) * increment));
	}

	//this is the price we pay for each pattern we index for a given region class
	map<region_attr_type, unsigned int> regionClassToBudget;
	map<region_attr_type, unsigned int> regionClassToNumRegions;
	map<region_attr_type, Vec<unsigned int> > regionClassToRegionLengths;

	//this is the markov model for each region class
	map<region_attr_type, TyphonMarkovModel> regionClassToMarkovModel;
	unsigned int numRegions = _regionList.getNumRegions();

	for (unsigned int i = 0; i < numRegions; i++)
	{
		region_attr_type type = _regionList.getRegionType(i);
		if (regionClassToBudget.find(type) == regionClassToBudget.end())
		{
			regionClassToBudget[type] = 0;
			regionClassToMarkovModel[type].setOrder(order);
			for (unsigned int j = 0; j < regionLengths.size(); ++j)
			{
				regionClassToRegionLengths[type].push_back(0);
			}
		}
		const Region& curRegion = _regionList.getRegion(i);
		regionClassToBudget[type] += curRegion.getLength();
		regionClassToNumRegions[type]++;
		//regionClassToMarkovModel[type].addCountsOf(pv, curRegion);
		bool found = false;
		for (unsigned int j = 0; j < regionLengths.size() - 1; ++j)
		{
			if (curRegion.getLength() < regionLengths[j])
			{
				regionClassToRegionLengths[type][j]++;
				found = true;
				break;
			}
		}
		if (!found)
		{
			regionClassToRegionLengths[type].back()++;
		}
	}

	map<region_attr_type, unsigned int> regionClassToAvgLength;
	map<region_attr_type, unsigned int>::iterator mEnd = regionClassToBudget.end();
	for (map<region_attr_type, unsigned int>::iterator mIt = regionClassToBudget.begin(); mIt != mEnd; ++mIt)
	{
		unsigned int avgLength = mIt->second / regionClassToNumRegions[mIt->first];
		regionClassToAvgLength[mIt->first] = avgLength > MAX_REGION_LENGTH ? MAX_REGION_LENGTH : avgLength;
	}

	Vec<region_attr_type> regionTypes;
	_regionList.getRegionTypes(regionTypes);
	
	//okay, we need to have a sorted list of region attr types to the gain in
	//number of regions if we pick the next pattern
	list<pair<region_attr_type, double> > gainPerRegionClass;
	//this is a mapping from region_attr_type to the next pattern we'll index
	map<region_attr_type, Vec<TyphonNS::Pattern>::size_type> regionClassToNextPattern;
	//this is a map from the region class to the expected numbe of regions covered
	//GIVEN that this region class exists
	map<region_attr_type, double> conditionalProbabilities;
	map<region_attr_type, Vec<double> > conditionalProbabilitiesForEachRegion;

	//get the best pattern for each region class
	const Vec<TyphonNS::Pattern>& patternCandidates = _getPatterns();

	Vec<region_attr_type>::iterator end = regionTypes.end();	
	for (Vec<region_attr_type>::iterator it = regionTypes.begin(); it != end; ++it)
	{
		regionClassToMarkovModel[*it].setPercentId(RegionList::getConsLevel(*it));

		//pair<double, Vec<TyphonNS::Pattern>::size_type> bestPatternPair = MandalaWrapper::bestPattern(regionClassToAvgLength[*it], regionClassToMarkovModel[*it], patternCandidates, _regionClassToPatternSubset[*it]);

		Vec<TyphonNS::Pattern>::size_type nextPatternToAdd = _regionClassToPatternSubset[*it].size();
		pair<double, Vec<TyphonNS::Pattern>::size_type> bestPatternPair(MandalaWrapper::evaluatePattern(regionClassToAvgLength[*it], regionClassToMarkovModel[*it], patternCandidates, nextPatternToAdd, _regionClassToPatternSubset[*it]), nextPatternToAdd);

		double initialConditionalProbability = bestPatternPair.first;

		double additionalExpectedNumberOfRegions = 0;
		if (_eachRegion)
		{
			//go through for each region
			Vec<double>& probs = conditionalProbabilitiesForEachRegion[*it];
			Vec<unsigned int>::size_type rlSize = regionLengths.size();
			for (unsigned int i = 0; i < rlSize; ++i)
			{
				double prob = MandalaWrapper::evaluatePattern(regionLengths[i], regionClassToMarkovModel[*it], patternCandidates, bestPatternPair.second, _regionClassToPatternSubset[*it]);
				probs.push_back(prob);
				additionalExpectedNumberOfRegions = RegionList::getWeight(*it) * prob * regionLengths[i] * regionClassToRegionLengths[*it][i];
			}
		}
		else
		{
			conditionalProbabilities[*it] = initialConditionalProbability;
			additionalExpectedNumberOfRegions = RegionList::getWeight(*it) * initialConditionalProbability * regionClassToNumRegions[*it];
		}

		pair<region_attr_type, double> newPair(*it, additionalExpectedNumberOfRegions / regionClassToBudget[*it]);
		gainPerRegionClass.insert(lower_bound(gainPerRegionClass.begin(), gainPerRegionClass.end(), newPair, &NextPatternSetGeneratorByRegionClassNS::compare), newPair);
		regionClassToNextPattern[*it] = bestPatternPair.second;
	}

	unsigned int budgetLeft = budget;
	//extract the minimum
	while (!gainPerRegionClass.empty())
	{
		list<pair<region_attr_type, double> >::iterator attrToAddToIt = gainPerRegionClass.begin();

		region_attr_type curType = attrToAddToIt->first;
		unsigned int budgetCost = regionClassToBudget[curType];

		gainPerRegionClass.erase(attrToAddToIt);

		if (budgetLeft > budgetCost || (budgetCost - budgetLeft) < budgetLeft)
		{
			if (budgetLeft > budgetCost)
			{
				budgetLeft -= budgetCost;
			}
			else
			{
				budgetLeft = 0;
			}

			_regionClassToPatternSubset[curType].push_back(regionClassToNextPattern[curType]);

			//cout << "Picked " << curType.first << "," << curType.second << " with cost " << budgetCost << "; has " << _regionClassToPatternSubset[curType].size() << "; " << conditionalProbabilities[curType] << "; budget left is " << budgetLeft << endl;

			if (_regionClassToPatternSubset[curType].size() < patternCandidates.size())
			{
				//pair<double, Vec<TyphonNS::Pattern>::size_type> bestPatternPair = MandalaWrapper::bestPattern(regionClassToAvgLength[curType], regionClassToMarkovModel[curType], patternCandidates, _regionClassToPatternSubset[curType]);

				Vec<TyphonNS::Pattern>::size_type nextPatternToAdd = _regionClassToPatternSubset[curType].size();
				pair<double, Vec<TyphonNS::Pattern>::size_type> bestPatternPair(MandalaWrapper::evaluatePattern(regionClassToAvgLength[curType], regionClassToMarkovModel[curType], patternCandidates, nextPatternToAdd, _regionClassToPatternSubset[curType]), nextPatternToAdd);
				
				double additionalExpectedNumberOfRegions = 0;
				if (_eachRegion)
				{
					//go through for each region
					Vec<double>& probs = conditionalProbabilitiesForEachRegion[curType];
					Vec<unsigned int>::size_type rlSize = regionLengths.size();
					for (unsigned int i = 0; i < rlSize; i++)
					{
						double prob = MandalaWrapper::evaluatePattern(regionLengths[i], regionClassToMarkovModel[curType], patternCandidates, bestPatternPair.second, _regionClassToPatternSubset[curType]);
						double additionalConditionalProbability = (1 - probs[i]) * prob;

						probs[i] += additionalConditionalProbability;
						additionalExpectedNumberOfRegions += RegionList::getWeight(curType) * additionalConditionalProbability * regionLengths[i] * regionClassToRegionLengths[curType][i];
					}
				}
				else
				{
					double additionalConditionalProbability = (1 - conditionalProbabilities[curType]) * bestPatternPair.first;
					conditionalProbabilities[curType] += additionalConditionalProbability;

					additionalExpectedNumberOfRegions = RegionList::getWeight(curType) * additionalConditionalProbability * regionClassToNumRegions[curType];
				}

				pair<region_attr_type, double> newPair(curType, additionalExpectedNumberOfRegions / regionClassToBudget[curType]);
				gainPerRegionClass.insert(lower_bound(gainPerRegionClass.begin(), gainPerRegionClass.end(), newPair, &NextPatternSetGeneratorByRegionClassNS::compare), newPair);
				regionClassToNextPattern[curType] = bestPatternPair.second;
			}
		}
	}

	double numRegionsExpecting = 0;
	for (Vec<region_attr_type>::iterator it = regionTypes.begin(); it != end; ++it)
	{
		if (_eachRegion)
		{

		}
		else
		{
			numRegionsExpecting += regionClassToNumRegions[*it] * conditionalProbabilities[*it] * RegionList::getWeight(*it);
		}
	}
}

bool NextPatternSetGeneratorByRegionClassNS::compare(const pair<NextPatternSetGeneratorByRegionClass::region_attr_type, double>& a, const pair<NextPatternSetGeneratorByRegionClass::region_attr_type, double>& b)
{
	return a.second > b.second;
}
