#include "NextPatternSetGeneratorOneRegion.h"
#include "MandalaWrapper.h"

void NextPatternSetGeneratorOneRegion::nextPatternSet(Region& r, Vec<Vec<TyphonNS::Pattern>::size_type>& p)
{
	const Vec<TyphonNS::Pattern>& patternCandidates = _getPatterns();

	//we need to 
	while (hasMoreBudget())
	{
		if (p.size() == patternCandidates.size())
		{
			_finished();
			break;
		}
		//pair<double, Vec<TyphonNS::Pattern>::size_type> bestPattern = MandalaWrapper::bestPattern(_length, _tmm, patternCandidates, p);
		Vec<TyphonNS::Pattern>::size_type nextPatternToAdd = p.size();
		pair<double, Vec<TyphonNS::Pattern>::size_type> bestPattern(MandalaWrapper::evaluatePattern(_length, _tmm, patternCandidates, nextPatternToAdd, p), nextPatternToAdd);
		
		p.push_back(bestPattern.second);
		_subtractFromBudget(_end - _start + 1);
	}
	r.setStart(_start);
	r.setEnd(_end);
}


void NextPatternSetGeneratorOneRegion::_initialize(const ProbVector& pv, const RegionList& r, const unsigned int budget, const unsigned int order)
{
	//we store the lowest and highest portion to index
	if (r.getNumRegions() == 0)
	{
		_finished();
	}
	else
	{
		_start = r.getRegion(0).getStart();
		_end = r.getRegion(r.getNumRegions() - 1).getEnd();

		//first construct the markov model
		_tmm.setOrder(order);

		unsigned int numRegions = r.getNumRegions();
		_length = 0;
		for (unsigned int i = 0; i < numRegions; i++)
		{
			Region curRegion = r.getRegion(i);
			_tmm.addCountsOf(pv, curRegion);
			_length += curRegion.getLength();
		}
		if (numRegions)
		{
			_length /= numRegions;
		}
		if (_length > MAX_REGION_LENGTH)
		{
			_length = MAX_REGION_LENGTH;
		}
	}
}
