#include "Pattern.h"
#include <sstream>

using namespace std;

TyphonNS::Pattern::Pattern(const Vec<bool> p) : _pattern(p), _kLength(0), _patternLength(_pattern.size())
{
	//calculate the number of ones
  for (size_type i = 0; i < _patternLength; i++)
  {
    if (_pattern[i])
    {
      _kLength++;
    }
  }
}

TyphonNS::Pattern::Pattern(const Vec<bool>::size_type w) : _pattern(w, true), _kLength(w), _patternLength(w)
{
}


TyphonNS::Pattern::Pattern() : _pattern(Vec<bool>(0)), _kLength(0), _patternLength(0)
{

}

TyphonNS::Pattern::Pattern(const char* const stringRep) : _kLength(0), _patternLength(0)
{
  _patternLength = strlen(stringRep);
  for (size_type i = 0; i < _patternLength; i++)
  {
    char temp[2];
    sprintf(temp, "%c", stringRep[i]);
    bool cur = atoi(temp);
    if (cur)
    {
      _kLength++;
    }
    _pattern.push_back(atoi(temp));
  }
}

TyphonNS::Pattern::Pattern(const unsigned int intRep, const TyphonNS::Pattern::size_type size) : _kLength(0), _patternLength(size)
{
	unsigned int copyOfIntRep = intRep;
	while (_pattern.size() < size)
	{
		bool lowestOrderBit = copyOfIntRep % 2;
		if (lowestOrderBit)
		{
			_kLength++;
		}
		_pattern.push_back(lowestOrderBit);
		copyOfIntRep >>= 1u;
	}
}

TyphonNS::Pattern::Pattern(const Vec<TyphonNS::Pattern::size_type>& zeroPositions, TyphonNS::Pattern::size_type size) : _pattern(size, true), _kLength(size - zeroPositions.size()), _patternLength(size)
{
	Vec<TyphonNS::Pattern::size_type>::const_iterator end = zeroPositions.end();
	for (Vec<TyphonNS::Pattern::size_type>::const_iterator it = zeroPositions.begin(); it != end; ++it)
	{
		if (*it < size)
		{
			_pattern[*it] = false;
		}
		else
		{
			_kLength++;
		}
	}

}

string TyphonNS::Pattern::toString() const
{
	ostringstream stringRep("");
  for (size_type i = 0; i < _patternLength; i++)
  {
    stringRep << _pattern[i];
  }
  return stringRep.str();
}

void TyphonNS::Pattern::trunc(TyphonNS::Pattern::size_type newSize)
{
	if (newSize > _kLength)
	{
		return;
	}
	else
	{
		Vec<bool> newPattern;
		TyphonNS::Pattern::size_type numOnesPushed = 0;
		for (TyphonNS::Pattern::size_type i = 0; numOnesPushed < newSize; i++)
		{
			if (_pattern[i]) 
			{
				numOnesPushed++;
			}
			newPattern.push_back(_pattern[i]);
		}

		_pattern.clear();
		_pattern = newPattern;
		_kLength = newSize;
		_patternLength = _pattern.size();
	}
}

void TyphonNS::Pattern::trim(TyphonNS::Pattern::size_type newSize)
{
	if (newSize > _patternLength)
	{
		return;
	}
	else
	{
		unsigned int newZeros = _kLength > newSize ? 0 : newSize - _kLength;
		unsigned int curZeros = 0;
		Vec<bool> newPattern;
		for (TyphonNS::Pattern::size_type i = 0; i < _patternLength && newPattern.size() < newSize; i++)
		{
			if (!_pattern[i])
			{
				curZeros++;
				if (curZeros > newZeros)
				{
					continue;
				}
			}
			newPattern.push_back(_pattern[i]);
		}

		_pattern.clear();
		_pattern = newPattern;
		_patternLength = _pattern.size();
	}
}


void TyphonNS::Pattern::pad(TyphonNS::Pattern::size_type numToAdd, bool value)
{
	_pattern.resize(_pattern.size() + numToAdd, value);
	_patternLength += numToAdd;
	if (value)
	{
		_kLength += numToAdd;
	}
}

bool TyphonNS::Pattern::operator[](size_type pos) const
{
  return _pattern[pos];
}

bool TyphonNS::Pattern::operator==(const Pattern& pattern) const
{
	if (_kLength != pattern._kLength)
	{
		return false;
	}
	else if (_patternLength != pattern._patternLength)
	{
		return false;
	}
	for (unsigned int i = 0; i < _patternLength; i++)
	{
		if (_pattern[i] != pattern._pattern[i])
		{
			return false;
		}
	}
	return true;
}

ostream& TyphonNS::operator<<(ostream& out, const TyphonNS::Pattern& p)
{
	out << p.toString();
	return out;
}	
