#ifndef _probvector_h_included_
#define _probvector_h_included_

#include "System.h"
#include "Seq.h"
#include "Mseq.h"
#include "DVector.h"
#include "Constants.h"
#include <string>
#include <iostream>

/**
 * A vector of tuples for probabilities
 */

class ProbVector
{
 public:

	typedef double prob_type;
	typedef Vec<prob_type>::size_type size_type;

	ProbVector() {};

	ProbVector(Mseq& m, double pc = 0);

	size_type size() const
		{
			if (_tuples.size() > 0)
			{
				return _tuples[0].size();
			}
			else
			{
				return 0;
			}
		}

	void resize(Seq::size_type length)
		{
			clear();
			_tuples.resize(BIO::NUM_ALPHABET + 1);
			for (DVector<DVector<prob_type> >::iterator it = _tuples.begin(); it != _tuples.end(); ++it)
			{
				it->resize(length, 0);
			}
		}

	void clear()
		{
			for (DVector<DVector<prob_type> >::iterator it = _tuples.begin(); it != _tuples.end(); ++it)
			{
				it->clear();
			}
		}

	void copyFrom(Seq::size_type to, const ProbVector& pv, Seq::size_type from)
		{
			for (unsigned int i = 0; i < _tuples.size(); i++)
			{
				_tuples[i][to] = pv._tuples[i][from];
			}
		}

	void setProbability(Seq::size_type pos, unsigned char character, prob_type prob)
		{
			_tuples[character][pos] = prob;
		}

	prob_type getProbability(Seq::size_type pos, unsigned char character) const
		{
			return _tuples[character][pos];
		}

	prob_type getPresentProbability(Seq::size_type pos) const
		{
			return 1 - _getGapProbability(pos);
		}

	void setPresentProbability(Seq::size_type pos, prob_type value) 
		{
			_tuples[GAP_INDEX][pos] = 1 - value;
		}

	void setMaxLetterProb(Seq::size_type pos, prob_type value) 
		{
			_tuples[BIO::NUM_ALPHABET][pos] = value;
		}

	prob_type getMaxLetterProb(Seq::size_type pos) const
		{
			if (_tuples[BIO::NUM_ALPHABET][pos])
			{
				return _tuples[BIO::NUM_ALPHABET][pos];
			}

			prob_type maxProb = 0;

			//for (DVector<DVector<prob_type> >::const_iterator it = _tuples.begin(); it != _tuples.end() - 2; ++it)
			for (unsigned int i = 0; i < BIO::NUM_NUCLEOTIDES; i++)
			{
				if (_tuples[i][pos] > maxProb)
				{
					maxProb = _tuples[i][pos];
				}
			}

			//prob_type gapProb = _getGapProbability(pos);
			//maxProb = gapProb == 1 ? 0 : maxProb / (1 - gapProb);
			
			return maxProb;
		}

	unsigned int getMaxLetter(Seq::size_type pos) const
		{
			prob_type maxProb = 0;
			unsigned int maxLetter = 0;
			//for (DVector<DVector<prob_type> >::const_iterator it = _tuples.begin(); it != _tuples.end() - 2; ++it, ++i)
			for (unsigned int i = 0; i < BIO::NUM_NUCLEOTIDES; i++)
			{
				if (_tuples[i][pos] > maxProb)
				{
					maxProb = _tuples[i][pos];
					maxLetter = i;
				}
			}
			return maxLetter;
		}

	const static string TAG;

	void readFromFile(const string& filename);
	void writeToFile(const string& filename);

 private:

	void _increment(Seq::size_type pos, unsigned char character, prob_type inc)
		{
			_tuples[character][pos] += inc;
		}

	void _normalize(Seq::size_type pos)
		{
			prob_type totalLetterProb = 0;
			for (unsigned int i = 0; i < BIO::NUM_NUCLEOTIDES; i++)
			{
				totalLetterProb += _tuples[i][pos];
			}
			prob_type totalGapProb = _tuples[BIO::NUM_NUCLEOTIDES][pos];

			for (unsigned int i = 0; i < BIO::NUM_NUCLEOTIDES; i++)
			{
				_tuples[i][pos] /= totalLetterProb;
			}
			_tuples[BIO::NUM_NUCLEOTIDES][pos] /= (totalLetterProb + totalGapProb);
		}

	prob_type _getGapProbability(Seq::size_type pos) const
		{
			return _tuples[GAP_INDEX][pos];
		}


	static const Vec<prob_type>::size_type GAP_INDEX = BIO::NUM_ALPHABET - 1;

	DVector<DVector<prob_type> > _tuples;

};

#endif

