#include "TyphonMarkovModel.h"

#include <deque>

void TyphonMarkovModel::addCountsOf(const ProbVector& pv, const Region& region)
{
	unsigned int curRegionBegin = region.getStart();
	unsigned int curRegionEnd = region.getEnd() + 1;
	unsigned int curRegionSize = curRegionBegin - curRegionEnd;
	if (curRegionSize < _order + 1)
	{
		return;
	}

	//rampup
	deque<double> curProbs;
	for (unsigned int i = curRegionBegin; i < curRegionBegin + _order + 1; ++i)
	{
		curProbs.push_back(pv.getMaxLetterProb(i));
	}

	//we need to initialize the counts 
	Vec<double> currentCounts(_counts.size());
	for (unsigned int i = 0; i < currentCounts.size(); ++i)
	{
		//get the probability
		double count = 1;
		for (unsigned int j = 0; j < _order + 1; j++)
		{
			double maxProb = *(curProbs.rbegin() + j);
			if ((i >> j) & 1)
			{
				count *= maxProb;
			}
			else
			{
				count *= (1 - maxProb);
			}
		}
		currentCounts[i] = count;
		_counts[i] += currentCounts[i];
	}

	//now main loop
	for (unsigned int i = curRegionBegin + _order + 1; i < curRegionEnd; ++i)
	{
		//first we need to drop off the first letter of every entry
		for (unsigned int j = currentCounts.size() / 2; j < currentCounts.size(); ++j)
		{
			//divide by the probability of a 1 at the beginning
			currentCounts[j] /= curProbs.front();
		}

		curProbs.pop_front();
		curProbs.push_back(pv.getMaxLetterProb(i));

		//now update
		unsigned int k = 0;
		for (unsigned int j = currentCounts.size() / 2; j < currentCounts.size(); ++j)
		{
			currentCounts[k] = currentCounts[j] * (1 - curProbs.back());
			_counts[k] += currentCounts[k];
			k++;
			currentCounts[k] = currentCounts[j] * curProbs.back();
			_counts[k] += currentCounts[k];
			k++;
		}
	}
}

void TyphonMarkovModel::projectTo(TyphonMarkovModel& m)
{
	if (m._order > _order)
	{
		return;
	}
	else
	{
		Vec<double> newCounts(_counts.size());
		copy(_counts.begin(), _counts.end(), newCounts.begin());
		while (newCounts.size() > m._counts.size())
		{
			Vec<double> tempCounts(newCounts.size() / 2, 0);
			for (Vec<double>::size_type i = 0; i < tempCounts.size(); i++)
			{
				tempCounts[i] = newCounts[i] + newCounts[i + tempCounts.size()];
			}
			newCounts.resize(tempCounts.size(), 0);
			copy(tempCounts.begin(), tempCounts.end(), newCounts.begin());		
		}
		copy(newCounts.begin(), newCounts.end(), m._counts.begin());		
	}
}

