#ifndef _typhonmarkovmodel_h_included_
#define _typhonmarkovmodel_h_included_

#include "System.h"
#include "Vec.h"
#include "ProbVector.h"
#include "Region.h"
#include <cmath>

class TyphonMarkovModel
{
 public:
	
	TyphonMarkovModel() {}

	TyphonMarkovModel(unsigned int order) : _order(order)
		{
			_counts.resize(static_cast<unsigned int>(pow(2.0, static_cast<double>(order + 1))), 0);
		}

	void setPercentId(double id)
		{
			if (id > 1)
			{
				id = 1;
			}
			setOrder(0);
			_counts[0] = (1 - id);
			_counts[1] = id;
		}


	void setOrder(unsigned int order)
		{
			_order = order; 
			_counts.clear();
			_counts.resize(static_cast<unsigned int>(pow(2.0, static_cast<double>(order + 1))), 0);
		}

	void addCountsOf(const ProbVector& pv, const Region& region);

	void projectTo(TyphonMarkovModel& m);

	unsigned int getOrder() { return _order; }

	unsigned int getNumHistories() { return _counts.size() / 2; }

	double getProb(unsigned int number) { return _counts[2 * number + 1] / (_counts[2 * number] + _counts[2 * number + 1]); }

 private:

	unsigned int _order;
	Vec<double> _counts;

};

#endif
